# coding: utf-8
# Author: Felix Fontein <tkuratom@redhat.com>
# Author: Toshio Kuratomi <tkuratom@redhat.com>
# License: GPLv3+
# Copyright: Ansible Project, 2020

import os
import os.path
import typing as t

from packaging.version import Version as PypiVer

from antsibull_changelog.changelog_generator import (
    ChangelogGenerator,
    ChangelogEntry as ChangelogGeneratorEntry,
)
from antsibull_changelog.config import DEFAULT_SECTIONS
from antsibull_changelog.rst import RstBuilder

from . import app_context
from .changelog import Changelog, ChangelogData, ChangelogEntry, CollectionsMetadata, get_changelog


#
# Changelog
#


PluginDataT = t.List[t.Tuple[str, str, ChangelogGenerator, t.Optional[ChangelogGeneratorEntry]]]


def _cleanup_plugins(entries: t.List[t.Any]) -> t.List[t.Any]:
    '''Remove duplicate module/plugin/object entries from the given list.
    '''
    result = []
    found_names = set()
    for entry in entries:
        name = entry['name']
        if name in found_names:
            continue
        result.append(entry)
        found_names.add(name)
    return result


def optimize_release_entry(entry: ChangelogGeneratorEntry) -> ChangelogGeneratorEntry:
    '''Remove duplicate entries from changelog entry.

    This can happen if entries from versions from different changelogs are combined
    which partially include each other's changes. This happens if for example the 1.1.0
    changelog continues the 1.0.2 changelog, and there was also a 1.0.3 release continuing
    the 1.0.2 changelog. The 1.0.3 changes are usually a subset of the 1.1.0 changes as
    they are backported bugfixes.
    '''
    entry.modules = _cleanup_plugins(entry.modules)
    for plugin_type, plugins in entry.plugins.items():
        entry.plugins[plugin_type] = _cleanup_plugins(plugins)
    for object_type, objects in entry.objects.items():
        entry.objects[object_type] = _cleanup_plugins(objects)
    for section, changes in entry.changes.items():
        if isinstance(changes, list):
            entry.changes[section] = sorted(set(changes))
    return entry


def _add_rst_table_row(builder: RstBuilder, column_widths: t.List[int], row: t.List[str]):
    lines = [[''] * len(column_widths)]
    for i, value in enumerate(row):
        for j, line in enumerate(value.splitlines()):
            if j >= len(lines):
                lines.append([''] * len(column_widths))
            lines[j][i] = line
    for line in lines:
        parts = ['|']
        for j, cell in enumerate(line):
            parts.append(' ' + cell + ' ' * (1 + column_widths[j] - len(cell)) + '|')
        builder.add_raw_rst(''.join(parts))


def _add_rst_table_line(builder: RstBuilder, column_widths: t.List[int], sep: str):
    parts = ['+']
    for w in column_widths:
        parts.append(sep * (w + 2) + '+')
    builder.add_raw_rst(''.join(parts))


def render_rst_table(builder: RstBuilder, headings: t.List[str],
                     cells: t.List[t.List[str]]) -> None:
    # Determine column widths
    column_widths = []
    for row in [headings] + cells:
        while len(row) > len(column_widths):
            column_widths.append(0)
        for i, value in enumerate(row):
            for line in value.splitlines():
                column_widths[i] = max(column_widths[i], len(line))

    # Add rows
    _add_rst_table_line(builder, column_widths, '-')
    _add_rst_table_row(builder, column_widths, headings)
    _add_rst_table_line(builder, column_widths, '=')
    for row in cells:
        _add_rst_table_row(builder, column_widths, row)
        _add_rst_table_line(builder, column_widths, '-')


def append_changelog_changes_collections(builder: RstBuilder,
                                         collection_metadata: CollectionsMetadata,
                                         changelog_entry: ChangelogEntry,
                                         is_last: bool) -> PluginDataT:
    result: PluginDataT = []

    if changelog_entry.changed_collections:
        builder.add_section('Included Collections' if is_last else 'Changed Collections', 1)
        builder.add_raw_rst(
            'If not mentioned explicitly, the changes are reported in the combined changelog'
            ' below.\n')
        headings = [
            'Collection',
            f'Ansible {changelog_entry.prev_version}',
            f'Ansible {changelog_entry.version_str}',
            'Notes'
        ]
        cells = []
        for (
                collector, collection_version, prev_collection_version
        ) in changelog_entry.changed_collections:
            row = [collector.collection, '', str(collection_version), '']
            if prev_collection_version is not None:
                row[1] = str(prev_collection_version)
            changelog = collector.changelog
            if changelog:
                release_entries = changelog.generator.collect(
                    squash=True,
                    after_version=prev_collection_version,
                    until_version=collection_version)
                if not release_entries:
                    row[-1] = "The collection did not have a changelog in this version."
                elif release_entries[0].empty:
                    row[-1] = "There are no changes recorded in the changelog."
                else:
                    result.append((
                        collector.collection,
                        f"{collector.collection}.",
                        changelog.generator,
                        optimize_release_entry(release_entries[0])))
            else:
                metadata = collection_metadata.get_meta(collector.collection)
                if metadata.changelog_url is not None:
                    row[-1] = (
                        "You can find the collection's changelog at"
                        f" `{metadata.changelog_url} <{metadata.changelog_url}>`_."
                    )
                else:
                    row[-1] = (
                        "Unfortunately, this collection does not provide changelog data in a"
                        " format that can be processed by the changelog generator."
                    )
            cells.append(row)
        render_rst_table(builder, headings, cells)
        builder.add_raw_rst('')

    return result


def append_changelog_changes_ansible(builder: RstBuilder,
                                     changelog_entry: ChangelogEntry) -> PluginDataT:
    changelog = changelog_entry.ansible_changelog

    release_entries = changelog.generator.collect(
        squash=True,
        after_version=str(changelog_entry.prev_version) if changelog_entry.prev_version else None,
        until_version=changelog_entry.version_str)

    if not release_entries:
        return []

    release_entry = optimize_release_entry(release_entries[0])

    release_summary = release_entry.changes.pop('release_summary', None)
    if release_summary:
        builder.add_section('Release Summary', 1)
        builder.add_raw_rst(t.cast(str, release_summary))
        builder.add_raw_rst('')

    if release_entry.empty:
        return []

    return [("", "", changelog.generator, release_entry)]


def append_changelog_changes_base(builder: RstBuilder,
                                  changelog_entry: ChangelogEntry) -> PluginDataT:
    base_name = 'Ansible-base'
    if (
            changelog_entry.base_collector.latest.major > 2 or
            changelog_entry.base_collector.latest.minor > 10
    ):
        base_name = 'Ansible-core'
    builder.add_section(base_name, 1)

    builder.add_raw_rst(f"Ansible {changelog_entry.version} contains {base_name} "
                        f"version {changelog_entry.ansible_base_version}.")
    if changelog_entry.prev_ansible_base_version:
        if changelog_entry.prev_ansible_base_version == changelog_entry.ansible_base_version:
            builder.add_raw_rst(f"This is the same version of {base_name} as in "
                                "the previous Ansible release.\n")
            return []

        builder.add_raw_rst(f"This is a newer version than version "
                            f"{changelog_entry.prev_ansible_base_version} contained in the "
                            f"previous Ansible release.\n")

    changelog = changelog_entry.base_collector.changelog
    if not changelog:
        return []

    release_entries = changelog.generator.collect(
        squash=True,
        after_version=changelog_entry.prev_ansible_base_version,
        until_version=changelog_entry.ansible_base_version)

    if not release_entries:
        builder.add_raw_rst(f"{base_name} did not have a changelog in this version.")
        return []

    release_entry = release_entries[0]

    if release_entry.empty:
        builder.add_raw_rst("There are no changes recorded in the changelog.")
        return []

    builder.add_raw_rst("The changes are reported in the combined changelog below.")
    return [(base_name, "ansible.builtin.", changelog.generator, release_entry)]


def common_start(a: t.List[t.Any], b: t.List[t.Any]) -> int:
    '''
    Given two sequences a and b, determines maximal index so that
    all elements up to that index are equal.
    '''
    common_len = min(len(a), len(b))
    for i in range(common_len):
        if a[i] != b[i]:
            return i
    return common_len


PluginDumpT = t.List[t.Tuple[t.List[str], str, str]]


def dump_items(builder: RstBuilder, items: PluginDumpT) -> None:
    last_title = []
    for title, name, description in sorted(items):
        if title != last_title:
            if last_title:
                builder.add_raw_rst('')
            for i in range(common_start(last_title, title), len(title)):
                builder.add_section(title[i], i + 1)
            last_title = title
        builder.add_list_item(f"{name} - {description}")

    if last_title:
        builder.add_raw_rst('')


def add_plugins(builder: RstBuilder, data: PluginDataT) -> None:
    plugins: PluginDumpT = []
    for name, prefix, dummy, release_entry in data:
        if release_entry:
            for plugin_type, plugin_datas in release_entry.plugins.items():
                for plugin_data in plugin_datas:
                    plugins.append((
                        # ['New Plugins', plugin_type.title(), name],
                        ['New Plugins', plugin_type.title()],
                        prefix + plugin_data['name'],
                        plugin_data['description']))
    dump_items(builder, plugins)


def add_objects(builder: RstBuilder, data: PluginDataT) -> None:
    objects: PluginDumpT = []
    for name, prefix, dummy, release_entry in data:
        if release_entry:
            for object_type, object_datas in release_entry.objects.items():
                for object_data in object_datas:
                    objects.append((
                        ['New {type}s'.format(type=object_type.title())],
                        prefix + object_data['name'],
                        object_data['description']))
    dump_items(builder, objects)


def add_modules(builder: RstBuilder, data: PluginDataT) -> None:
    modules: PluginDumpT = []
    for name, prefix, dummy, release_entry in data:
        if release_entry:
            for module in release_entry.modules:
                namespace = module.get('namespace') or ''
                if namespace.startswith('.ansible.collections.ansible_collections.'):
                    # Work around old antsibull-changelog versions which suffer from
                    # https://github.com/ansible-community/antsibull-changelog/issues/18
                    namespace = ''
                namespace = namespace.strip('.').split('.', 1) if namespace else []
                modules.append((
                    ['New Modules', name] + [ns.replace('_', ' ').title() for ns in namespace],
                    prefix + module['name'],
                    module['description']))
    dump_items(builder, modules)


def create_title_adder(builder: RstBuilder, title: str,
                       level: int) -> t.Generator[None, None, None]:
    builder.add_section(title, level)
    while True:
        yield


def append_removed_collections(builder: RstBuilder, changelog_entry: ChangelogEntry) -> None:
    if changelog_entry.removed_collections:
        builder.add_section('Removed Collections', 1)
        for collector, collection_version in changelog_entry.removed_collections:
            builder.add_list_item(f"{collector.collection} "
                                  f"(previously included version: {collection_version})")
        builder.add_raw_rst('')


def append_added_collections(builder: RstBuilder, changelog_entry: ChangelogEntry) -> None:
    if changelog_entry.added_collections:
        builder.add_section('Added Collections', 1)
        for collector, collection_version in changelog_entry.added_collections:
            builder.add_list_item(f"{collector.collection} (version {collection_version})")
        builder.add_raw_rst('')


def append_unchanged_collections(builder: RstBuilder, changelog_entry: ChangelogEntry) -> None:
    if changelog_entry.unchanged_collections:
        builder.add_section('Unchanged Collections', 1)
        for collector, collection_version in changelog_entry.unchanged_collections:
            builder.add_list_item(f"{collector.collection} (still version {collection_version})")
        builder.add_raw_rst('')


def append_changelog(builder: RstBuilder,
                     collection_metadata: CollectionsMetadata,
                     changelog_entry: ChangelogEntry,
                     is_last: bool) -> None:
    builder.add_section('v{0}'.format(changelog_entry.version_str), 0)

    builder.add_raw_rst('.. contents::')
    builder.add_raw_rst('  :local:')
    builder.add_raw_rst('  :depth: 2\n')

    # Add release summary for Ansible
    data = append_changelog_changes_ansible(builder, changelog_entry)

    append_removed_collections(builder, changelog_entry)
    append_added_collections(builder, changelog_entry)

    # Adds Ansible-base section
    data.extend(append_changelog_changes_base(builder, changelog_entry))
    builder.add_raw_rst('')

    # Adds list of changed collections
    data.extend(
        append_changelog_changes_collections(
            builder, collection_metadata, changelog_entry, is_last=is_last))

    # Adds all changes
    for section, section_title in DEFAULT_SECTIONS:
        maybe_add_section_title = create_title_adder(builder, section_title, 1)

        for name, dummy, dummy, release_entry in data:
            if not release_entry or release_entry.has_no_changes([section]):
                continue

            next(maybe_add_section_title)
            if name:
                builder.add_section(name, 2)
            release_entry.add_section_content(builder, section)
            builder.add_raw_rst('')

    # Adds new plugins and modules
    add_plugins(builder, data)
    add_modules(builder, data)
    add_objects(builder, data)

    # Adds list of unchanged collections
    append_unchanged_collections(builder, changelog_entry)


#
# Porting Guide
#


def append_porting_guide_section(builder: RstBuilder, changelog_entry: ChangelogEntry,
                                 maybe_add_title: t.Generator[None, None, None],
                                 section: str) -> None:
    maybe_add_section_title = create_title_adder(builder, section.replace('_', ' ').title(), 1)

    def check_changelog(
            name: str,
            changelog: t.Optional[ChangelogData],
            version: str,
            prev_version: t.Optional[str]) -> None:
        if not changelog:
            return
        entries = changelog.generator.collect(
            squash=True, after_version=prev_version, until_version=version)
        if not entries or entries[0].has_no_changes([section]):
            return
        next(maybe_add_title)
        next(maybe_add_section_title)
        if name:
            builder.add_section(name, 2)
        optimize_release_entry(entries[0]).add_section_content(builder, section)
        builder.add_raw_rst('')

    check_changelog(
        '',
        changelog_entry.ansible_changelog,
        changelog_entry.version_str,
        str(changelog_entry.prev_version) if changelog_entry.prev_version else None)
    is_core = (
        changelog_entry.base_collector.latest.major > 2 or
        changelog_entry.base_collector.latest.minor > 10
    )
    check_changelog(
        'Ansible-core' if is_core else 'Ansible-base',
        changelog_entry.base_collector.changelog,
        changelog_entry.ansible_base_version,
        changelog_entry.prev_ansible_base_version)
    for (
            collector, collection_version, prev_collection_version
    ) in changelog_entry.changed_collections:
        check_changelog(
            collector.collection,
            collector.changelog,
            collection_version,
            prev_collection_version)


def append_porting_guide(builder: RstBuilder, changelog_entry: ChangelogEntry) -> None:
    maybe_add_title = create_title_adder(
        builder, 'Porting Guide for v{0}'.format(changelog_entry.version_str), 0)

    for section in ['known_issues', 'breaking_changes', 'major_changes']:
        append_porting_guide_section(builder, changelog_entry, maybe_add_title, section)

    if changelog_entry.removed_collections:
        next(maybe_add_title)
        builder.add_section('Removed Collections', 1)
        for collector, collection_version in changelog_entry.removed_collections:
            builder.add_list_item(f"{collector.collection} "
                                  f"(previously included version: {collection_version})")
        builder.add_raw_rst('')

    for section in ['removed_features', 'deprecated_features']:
        append_porting_guide_section(builder, changelog_entry, maybe_add_title, section)


def insert_after_heading(lines: t.List[str], content: str) -> None:
    has_heading = False
    for index, line in enumerate(lines):
        if line.startswith('***') and line == '*' * len(line):
            has_heading = True
        elif has_heading:
            if line:
                has_heading = False
            else:
                # First empty line after top-level heading: insert TOC
                lines.insert(index, content)
                return


#
# Release Notes
#


class ReleaseNotes:
    changelog_filename: str
    changelog_bytes: bytes

    porting_guide_filename: str
    porting_guide_bytes: bytes

    def __init__(self, changelog_filename: str, changelog_bytes: bytes,
                 porting_guide_filename: str, porting_guide_bytes: bytes):
        self.changelog_filename = changelog_filename
        self.changelog_bytes = changelog_bytes

        self.porting_guide_filename = porting_guide_filename
        self.porting_guide_bytes = porting_guide_bytes

    @staticmethod
    def _get_changelog_bytes(changelog: Changelog) -> bytes:
        builder = RstBuilder()
        if changelog.ansible_version.major > 2:
            version = f"{changelog.ansible_version.major}"
        else:
            version = f"{changelog.ansible_version.major}.{changelog.ansible_version.minor}"
        builder.set_title(f"Ansible {version} Release Notes")

        if changelog.ansible_ancestor_version:
            builder.add_raw_rst(
                f"This changelog describes changes since"
                f" Ansible {changelog.ansible_ancestor_version}.\n"
            )

        builder.add_raw_rst('.. contents::\n  :local:\n  :depth: 2\n')

        entries = [entry for entry in changelog.entries if not entry.is_ancestor]
        for index, changelog_entry in enumerate(entries):
            append_changelog(
                builder,
                changelog.collection_metadata,
                changelog_entry,
                is_last=index + 1 == len(entries))

        return builder.generate().encode('utf-8')

    @staticmethod
    def _append_base_porting_guide_bytes(builder: RstBuilder, changelog: Changelog) -> None:
        base_porting_guide = changelog.base_collector.porting_guide
        if base_porting_guide:
            lines = base_porting_guide.decode('utf-8').splitlines()
            lines.append('')
            found_topics = False
            found_empty = False
            for line in lines:
                if not found_topics:
                    if line.startswith('.. contents::'):
                        found_topics = True
                    continue
                if not found_empty:
                    if line == '':
                        found_empty = True
                    continue
                builder.add_raw_rst(line)
            if not found_empty:
                print('WARNING: cannot find TOC of ansible-base porting guide!')

    @staticmethod
    def _get_porting_guide_bytes(changelog: Changelog) -> bytes:
        if changelog.ansible_version.major > 2:
            version = f"{changelog.ansible_version.major}"
            base_version = changelog.base_collector.latest
            base_version = f"{base_version.major}.{base_version.minor}"
        else:
            version = f"{changelog.ansible_version.major}.{changelog.ansible_version.minor}"
            base_version = f"{changelog.ansible_version.major}.{changelog.ansible_version.minor}"
        builder = RstBuilder()
        builder.add_raw_rst(
            f"..\n"
            f"   THIS DOCUMENT IS AUTOMATICALLY GENERATED BY ANTSIBULL! PLEASE DO NOT EDIT"
            f" MANUALLY! (YOU PROBABLY WANT TO EDIT porting_guide_base_{base_version}.rst)\n")
        builder.add_raw_rst(f".. _porting_{version}_guide:\n")
        builder.set_title(f"Ansible {version} Porting Guide")

        if changelog.ansible_version.major == 2:
            builder.add_raw_rst(
                # noqa: E501
                ".. warning::"
                "\n\n"
                "        "
                " In Ansible 2.10, many plugins and modules have migrated to"
                " Collections on `Ansible Galaxy <https://galaxy.ansible.com>`_. Your playbooks"
                " should continue to work without any changes. We recommend you start using the"
                " fully-qualified collection name (FQCN) in your playbooks as the explicit and"
                " authoritative indicator of which collection to use as some collections may"
                " contain duplicate module names."
                " You can search the `index of all modules"
                " <https://docs.ansible.com/ansible/2.10/collections/index_module.html>`_"
                " to find the collection a module has been relocated to."
                "\n\n"
                "This section discusses the behavioral changes between Ansible 2.9 and"
                " Ansible 2.10."
                "\n\n"
                "It is intended to assist in updating your playbooks, plugins and other parts of"
                " your Ansible infrastructure so they will work with this version of Ansible."
                "\n\n"
                "We suggest you read this page along with the `Ansible 2.10 Changelog <https://"
                "github.com/ansible-community/ansible-build-data/blob/main/2.10/"
                "CHANGELOG-v2.10.rst>`_"
                " to understand what updates you may need to make."
                "\n\n"
                "Since 2.10, Ansible consists of two parts:\n\n"
                "* ansible-base, which includes the command line tools with a small selection"
                " of plugins and modules, and\n"
                "* a `set of collections <https://github.com/ansible-community/ansible-build-data/"
                "blob/main/2.10/ansible.in>`_."
                "\n\n"
                "The :ref:`porting_2.10_guide_base` is included in this porting guide. The complete"
                " list of porting guides can be found at :ref:`porting guides <porting_guides>`."
                "\n"
            )

        builder.add_raw_rst('.. contents::\n  :local:\n  :depth: 2\n')

        # Determine ansible-base/-core version in previous major release
        prev_base_version = ''
        if any(entry.is_ancestor for entry in changelog.entries):
            # If there is an ancestor, the earliest ansible-base/-core version will be the
            # version used in the previous major release.
            prev_base_version = changelog.base_collector.earliest
            prev_base_version = f"{prev_base_version.major}.{prev_base_version.minor}"

        # Determine whether to include ansible-base/-core porting guide or not
        if changelog.ansible_version.major == 2 or base_version != prev_base_version:
            builder.add_raw_rst(
                # noqa: E501
                "\n"
                f"We suggest you read this page along with the `Ansible {version} Changelog"
                f" <https://github.com/ansible-community/ansible-build-data/blob/main/{version}/"
                f"CHANGELOG-v{version}.rst>`_ to understand what updates you may need to make."
                "\n")
            ReleaseNotes._append_base_porting_guide_bytes(builder, changelog)
        elif changelog.ansible_version.major == 3:
            # Special message for Ansible 3
            builder.add_raw_rst(
                # noqa: E501
                "\n"
                "Ansible 3 is based on Ansible-Base 2.10, which is the same major release as"
                " Ansible 2.10.  Therefore, there is no section on ansible-base in this porting"
                " guide.  If you are upgrading from Ansible 2.9, please first consult the"
                " Ansible 2.10 porting guide before continuing with the Ansible 3 porting guide."
                "\n\n"
                "We suggest you read this page along with the `Ansible 3 Changelog"
                " <https://github.com/ansible-community/ansible-build-data/blob/main/3/"
                "CHANGELOG-v3.rst>`_ to understand what updates you may need to make."
                "\n"
            )
        else:
            # Generic message if we again have two consecutive versions with the same ansible-core
            prev_version = changelog.ansible_version.major - 1
            prev_prev_version = changelog.ansible_version.major - 2
            builder.add_raw_rst(
                # noqa: E501
                "\n"
                f"Ansible {version} is based on Ansible-core {base_version}, which is the same"
                f" major release as Ansible {prev_version}.  Therefore, there is no section on"
                " ansible-core in this porting guide.  If you are upgrading from Ansible"
                f" {prev_prev_version}, please first consult the Ansible {prev_version} porting"
                f" guide before continuing with the Ansible {version} porting guide."
                "\n\n"
                f"We suggest you read this page along with the `Ansible {version} Changelog"
                f" <https://github.com/ansible-community/ansible-build-data/blob/main/{version}/"
                f"CHANGELOG-v{version}.rst>`_ to understand what updates you may need to make."
                "\n"
            )

        for porting_guide_entry in changelog.entries:
            if not porting_guide_entry.is_ancestor:
                append_porting_guide(builder, porting_guide_entry)

        return builder.generate().encode('utf-8')

    @classmethod
    def build(cls, changelog: Changelog) -> 'ReleaseNotes':
        if changelog.ansible_version.major > 2:
            version = f"{changelog.ansible_version.major}"
        else:
            version = f"{changelog.ansible_version.major}.{changelog.ansible_version.minor}"
        return cls(
            f"CHANGELOG-v{version}.rst",
            cls._get_changelog_bytes(changelog),
            f"porting_guide_{version}.rst",
            cls._get_porting_guide_bytes(changelog),
        )

    def write_changelog_to(self, dest_dir: str) -> None:
        path = os.path.join(dest_dir, self.changelog_filename)
        with open(path, 'wb') as changelog_fd:
            changelog_fd.write(self.changelog_bytes)

    def write_porting_guide_to(self, dest_dir: str) -> None:
        path = os.path.join(dest_dir, self.porting_guide_filename)
        with open(path, 'wb') as porting_guide_fd:
            porting_guide_fd.write(self.porting_guide_bytes)


def build_changelog() -> int:
    '''Create changelog and porting guide CLI command.'''
    app_ctx = app_context.app_ctx.get()

    ansible_version: PypiVer = app_ctx.extra['ansible_version']
    data_dir: str = app_ctx.extra['data_dir']
    dest_data_dir: str = app_ctx.extra['dest_data_dir']
    collection_cache: t.Optional[str] = app_ctx.extra['collection_cache']

    changelog = get_changelog(ansible_version, deps_dir=data_dir, collection_cache=collection_cache)

    release_notes = ReleaseNotes.build(changelog)
    release_notes.write_changelog_to(dest_data_dir)
    release_notes.write_porting_guide_to(dest_data_dir)

    missing_changelogs = []
    last_entry = changelog.entries[0]
    last_version_collectors = [
        collector for collector in changelog.collection_collectors
        if last_entry.version in last_entry.versions_per_collection[collector.collection]
    ]
    for collector in last_version_collectors:
        if collector.changelog is None:
            missing_changelogs.append(collector.collection)
    if missing_changelogs:
        print(f"{len(missing_changelogs)} out of {len(last_version_collectors)} collections"
              f" have no compatible changelog:")
        for collection_name in missing_changelogs:
            meta = changelog.collection_metadata.get_meta(collection_name)
            entry = [collection_name]
            if meta.changelog_url:
                entry.append(f"(changelog URL: {meta.changelog_url})")
            print(f"    {'  '.join(entry)}")
    return 0



# coding: utf-8
# Author: Toshio Kuratomi <tkuratom@redhat.com>
# License: GPLv3+
# Copyright: Toshio Kuratomi, 2021

import typing as t

import pydantic as p

from .config import DEFAULT_LOGGING_CONFIG, LoggingModel
from .validators import convert_bool, convert_none
from ..utils.collections import ContextDict


class BaseModel(p.BaseModel):
    """
    Configuration for all Context object classes.

    :cvar Config: Sets the following information

        :cvar allow_mutation: ``False``.  Prevents setattr on the contexts.
        :cvar extra: ``p.Extra.forbid``.  Prevents extra fields on the contexts.
        :cvar validate_all: ``True``.  Validates default values as well as user supplied ones.
    """

    class Config:
        """
        Set default configuration for building the context models.

        :cvar allow_mutation: ``False``.  Prevents setattr on the contexts.
        :cvar extra: ``p.Extra.forbid``.  Prevents extra fields on the contexts.
        :cvar validate_all: ``True``.  Validates default values as well as user supplied ones.
        """

        allow_mutation = False
        extra = p.Extra.forbid
        validate_all = True


class AppContext(BaseModel):
    """
    Structure and defaults of the app_ctx.

    :ivar extra: a mapping of arg/config keys to values.  Anything in here is unchecked by a
        schema.  These are usually leftover command line arguments and config entries. If
        values stored in extras need default values, they need to be set outside of the context
        or the entries can be given an actual entry in the AppContext to take advantage of the
        schema's checking, normalization, and default setting.
    :ivar ansible_base_url: Url to the ansible-core git repo.
    :ivar breadcrumbs: If True, build with breadcrumbs on the plugin pages (this takes more memory).
    :ivar galaxy_url: URL of the galaxy server to get collection info from
    :ivar indexes: If True, create index pages for all collections and all plugins in a collection.
    :ivar logging_cfg: Configuration of the application logging
    :ivar pypi_url: URL of the pypi server to query for information
    """

    extra: ContextDict = ContextDict()
    # pyre-ignore[8]: https://github.com/samuelcolvin/pydantic/issues/1684
    ansible_base_url: p.HttpUrl = 'https://github.com/ansible/ansible/'
    breadcrumbs: p.StrictBool = True
    # pyre-ignore[8]: https://github.com/samuelcolvin/pydantic/issues/1684
    galaxy_url: p.HttpUrl = 'https://galaxy.ansible.com/'
    indexes: p.StrictBool = True
    logging_cfg: LoggingModel = LoggingModel.parse_obj(DEFAULT_LOGGING_CONFIG)
    # pyre-ignore[8]: https://github.com/samuelcolvin/pydantic/issues/1684
    pypi_url: p.HttpUrl = 'https://pypi.org/'

    _convert_bools = p.validator('breadcrumbs', 'indexes',
                                 pre=True, allow_reuse=True)(convert_bool)


class LibContext(BaseModel):
    """
    Structure and defaults of the lib_ctx.

    :ivar chunksize: number of bytes to read or write at one time for network or file IO
    :ivar process_max: Maximum number of worker processes for parallel operations.  It may be None
        to mean, use all available CPU cores.
    :ivar thread_max: Maximum number of helper threads for parallel operations
    :ivar max_retries: Maximum number of times to retry an http request (in case of timeouts and
        other transient problems.
    :ivar doc_parsing_backend: The backend to use for parsing the documentation strings from
        plugins.  'ansible-internal' is the fastest.  'ansible-doc' exists in case of problems with
        the ansible-internal backend.
    """

    chunksize: int = 4096
    doc_parsing_backend: str = 'ansible-internal'
    max_retries: int = 10
    process_max: t.Optional[int] = None
    thread_max: int = 64

    _convert_nones = p.validator('process_max', pre=True, allow_reuse=True)(convert_none)



# Author: Toshio Kuratomi <tkuratom@redhat.com>
# Author: Felix Fontein <felix@fontein.de>
# License: GPLv3+
# Copyright: Ansible Project, 2020

from ... import app_context
from ...logging import log
from ...venv import FakeVenvRunner
from .stable import generate_docs_for_all_collections


mlog = log.fields(mod=__name__)


def generate_current_docs(indexes: bool, squash_hierarchy: bool) -> int:
    flog = mlog.fields(func='generate_current_docs')
    flog.debug('Begin processing docs')

    app_ctx = app_context.app_ctx.get()

    venv = FakeVenvRunner()

    generate_docs_for_all_collections(
        venv, None, app_ctx.extra['dest_dir'], app_ctx.extra['collections'],
        create_indexes=indexes and not squash_hierarchy,
        squash_hierarchy=squash_hierarchy,
        breadcrumbs=app_ctx.breadcrumbs)

    return 0


def generate_docs() -> int:
    """
    Create documentation for the current-collection subcommand.

    Current collection documentation creates documentation for one or multiple currently
    installed collections.

    :arg args: The parsed comand line args.
    :returns: A return code for the program.  See :func:`antsibull.cli.antsibull_docs.main` for
        details on what each code means.
    """
    app_ctx = app_context.app_ctx.get()

    indexes: bool = app_ctx.indexes
    squash_hierarchy: bool = app_ctx.extra['squash_hierarchy']

    if app_ctx.extra['use_current']:
        return generate_current_docs(indexes, squash_hierarchy)

    raise NotImplementedError('Priority to implement subcommands is stable, devel, plugin, and'
                              ' then collection commands. Only --use-current is implemented'
                              ' for the collection subcommand right now.')



# coding: utf-8
# Author: Toshio Kuratomi <tkuratom@redhat.com>
# License: GPLv3+
# Copyright: Ansible Project, 2020
"""
The antsibull logging framework.

Logging should behave differently in a library versus an application setting.  As a library, the
logging messages should largely avoid assuming that the log has been configured sanely to display
messages.  That way a developer is not forced to understand and configure the log to start playing
with the library.

Once the library is used in an application, however, the application should give the user the
ability to configure the log.  At that point, the log should obey the user's choices.  We achieve
both of those goals by configuring the log for library output when the log is imported and then
initializing it for application usage when a set of functions are called.

.. seealso::

    * :mod:`antsibull.config` to see how the antsibull scripts allow user defined configuration
        to configure logging after the bootstrap phase is over.  This is the primary way that end
        users interact with the logging subsystem.


======================================
Logging from the library point of view
======================================

Merely importing this file sets up logging to be used as a library.  Essentially, this means that
logging output is disabled.  That way the library doesn't spam the user's screen with log messages.

An application that wishes to use the log must import the log and then call
antsibull.logging.initialize_app_logging() before any other parts of antsibull are imported.  See
the :ref:`Application Logging <application_logging>`_ section for more details.


Usage within a module
=====================

Our convention for logging with twiggy is that the name field reflects the Python package that the
code is coming from (in this case, it is already set to ``antsibull`` by :mod:`antsibull.logging`.)
At the toplevel of a module, set up a logger which has a field named ``mod`` which reflects the
module name:

.. code-block:: python

    # The antsibull log object with the name already set to `antsibull`.
    from logging import log

    # mlog stands for module log.  It's our convention to create a logger from the
    # antsibull.logging.log object in each module.  `fields()` takes an arbitrary set of keyword
    # args and returns a new log object.  Any log messages we emit with this log object (or its
    # children) will include the fields which were set on it.  Our convention is to create mlog with
    # `fields(mod=__name__)` so that messages we make from mlog (or its children) have a field named
    # `mod` containing the name of the module.

    # `mod` and the value set to the module name.
    mlog = log.fields(mod=__name__)

    TRICKY_COMPUTED_GLOBAL = [a**a for a in range(1, 4)]
    # Use mlog for logging interesting things that happen at the module level.  Notice that we send
    # the value of TRICKY_COMPUTED_GLOBAL in fields() rather than in debug().  Most twiggy emitters
    # format fields so that they're machine parsable whereas the message is freeform text.  Using
    # fields for things that make sense to be viewed as key=value allows for human readable but
    # machine parsable structured output.  For instance, this is the default output format:
    # DEBUG:antsibull:computed_global=[1, 4, 27]:mod=__main__|globals computed at module level
    mlog.fields(computed_global=TRICKY_COMPUTED_GLOBAL).debug('globals computed at module level')

Inside of a function, you should further refine this by creating a new log object with the name of
the function in the ``func`` field.  If the function is a method of an object, include the class
name so that you can track down the proper function if there are multiple classes with the same
method names:

.. code-block:: python

def test_function(argument1):
    # flog stands for function log.  It's our convention to use this name.
    # Create a new one in any function you want to log from.
    # By creating this from mlog, we copy any fields and other settings that we made to mlog.
    # Our convention is to use the `func` field to hold the name of the function we're in.
    flog = mlog.fields(func='test_function')

    # This would output:
    # DEBUG:antsibull:func=test_function:mod=__main__|Enter
    flog.debug('Enter')
    value = do_something(argument1)

    flog.debug('Leave')

class FooBar:
    def __init__(self):
        flog = mlog.fields(func='FooBar.__init__')
        flog.debug('Enter')

        self._x = initialize_x()
        self._y = initialize_y()

        self.position = self.calculate_position(self._x, self._y)

        flog.debug('Leave')


.. _logging_levels::

Logging levels
==============

Twiggy comes with several logging levels.  This is what they are and how the antsibull library uses
them.  It can help you choose which log level is appropriate to use in your application by default.

:CRITICAL: An error occurred that needs immediate user attention.  The system cannot or should not
    continue if something critical happened.  Example: Unable to write output due to filesystem
    permissions.
:ERROR: An error occurred that could be fixed by the user in the future.  For instance, if the
    purpose is to parse multiple csv files and enter them into a database and we find that a single
    file won't parse as csv, this may be recorded as an error if the system processes the other csv
    files and enters them.
:WARNING: Something happened that was unexpected to the program but the user has a choice
    whether to address or not.  For instance, if the program fails to find a config file
    in an expected system location but it will operate from hardcoded defaults in that case, this
    would be a warning.
:NOTICE: Something significant happened that would be important information for an end user.
    For instance, in a data processing pipeline, this could be used to record when each **major**
    step of the pipeline finished.
:INFO: Something happened which could be interesting when a user of the system is trying to
    figure out why or what their commands are doing.  For example, a user may wonder why
    they are getting a file not found error on a config file they specified at the command
    line.  Knowing the filepath that the system is using when it tries to open the path
    will help them know that.
:DEBUG: This is similar to ``INFO`` but only helpful for a programmer developing the code.  For
    example, information on functions being called (to trace through the execution path) would be
    appropriate here.  Also relevant would be recording internal state of important variables or
    (occassionally) dumping data being operated upon.


.. _application_logging::

==========================================
Logging from the application point of view
==========================================


Logging Setup
=============

An antsibull command (:file:`antsibull/cli/*.py`) should import the ``log`` object from this module.
The log object will be configured for use within the library at first (silent) so the application
should call :func:`antsibull.logging.initialize_app_logging` as soon as possible to tell the ``log``
that it is okay to emit messages.

The initial application logging configuration will log to stderr at the ``WARNING`` level or
higher.  If the :envvar:`ANTIBULL_EARLY_DEBUG` environment variable is set, then  it will log at
the ``DEBUG`` level rather than ``WARNING``.

The antsibull command should read the configuration settings, which may include user specified
logging configuration and application defaults, and then call :twiggy:func:`twiggy.dict_config` to
finish the setup.  At that point, logging calls will emit logs according to the user's
configuration.

Here's a sample of the relevant portions of an antsibull command to show how this will look:

.. code-block:: python

    # File is antsibull/cli/antsibull_command.py
    import twiggy
    # log is the toplevel log object.  It is important to import this and initialize it prior to
    # using the log so that sane defaults can be set.
    from ..logging import log, initialize_app_logging

    # By default, the log is configured to be useful within a library where the user may not have
    # been given the chance to configure the log.  Calling initialize_app_logging() reconfigures
    # the log to a more verbose state which the user can then configure.  This must be done before
    # any antsibull modules which setup their own loggers ( mlog = log.fields(__name__)) happen as
    # each of those loggers will copy the settings that log has at the time they are created.
    initialize_app_logging()

    from ..config import load_config


    mlog = log.fields(mod=__name__)

    def run(args):
        flog = mlog.fields(func='run')
        args = parse_args(args[0], args[1:])
        cfg = load_config(args.config_file)

        context_data = app_context.create_contexts(args=args, cfg=cfg)
        with app_context.app_and_lib_context(context_data) as (app_ctx, dummy_):
            # initialize_app_logging() sets the log's configuration with defaults appropriate for
            # an application but this call takes that one step further. It takes the logging
            # configuration from the user's config file and hands it to twiggy.dict_config() so
            # that the user has ultimate control over what log level, what format, and which file
            # the log is output as.  See the twiggy documentation for information on the format of
            # the logging config.  See the antsibull.app_context documentation if you want more
            # information on the context object.
            twiggy.dict_config(app_ctx.logging_cfg.dict())


Once those steps are taken, any further logging calls will obey the user's configuration.

"""
import os

import twiggy
import twiggy.levels

#: The standard log to use everywhere.  The name of the logger for all of the antsibull libraries
#: is antsibull so that it is easy to setup an emitter for all of antsibull.  For those used to
#: using the module's __name__ field as the name, the idiom we use here is to set the module name
#: in the ``mod`` field.
log = twiggy.log.name('antsibull').trace()  # pyre-ignore[16]: twiggy generates log dynamically

# We disable logging at the library level so that we don't unnecessarily spam the user with output.
# Applications which use the library should re-enable this and give the user the ability to control
# what output gets logged.  See the antisbull.logging docstring for more information.
log.min_level = twiggy.levels.DISABLED

mlog = log.fields(mod=__name__)
mlog.debug('logging loaded')


def plugin_filter():
    """
    Filter out messages which come from plugin error output.

    :arg msg: A :twiggy:obj:`twiggy.message.Message` object which would be filtered
    """
    def wrapped(msg):
        return (
            msg.fields['func'] == 'write_plugin_rst' and
            msg.fields['mod'] == 'antsibull.write_docs'
        )
    return wrapped


def initialize_app_logging():
    """
    Change log settings to make sense for an application.

    Merely importing the :mod:`antsibull.logging` module sets up the logger for use as part of
    a library.  Calling this function will initialize the logger for use in an application.
    """
    # We want to see logs from the antsibull library, so the very first thing we do is turn the log
    # level up to DEBUG.  That way individual emitters will be able to control the log level.
    log.min_level = twiggy.levels.DEBUG

    # Temporarily setup logging with defaults until we can get configuration from the user.
    _level = twiggy.levels.WARNING
    if os.environ.get('ANTSIBULL_EARLY_DEBUG', False):
        _level = twiggy.levels.DEBUG
    twiggy.quick_setup(min_level=_level)


__all__ = ('log', 'initialize_app_logging', 'plugin_filter')



# coding: utf-8
# Author: Felix Fontein <felix@fontein.de>
# License: GPLv3+
# Copyright: Ansible Project, 2021
"""
Functions for parsing and interpreting collection metadata.
"""

from collections import defaultdict

import datetime
import os
import typing as t

import asyncio
import asyncio_pool

from .. import app_context
from .. import yaml
from ..constants import DOCUMENTABLE_PLUGINS
from ..utils.collections import compare_all_but
from ..utils.get_pkg_data import get_antsibull_data
from . import AnsibleCollectionMetadata
from .fqcn import get_fqcn_parts


# A nested structure as follows:
#       plugin_type:
#           plugin_name:  # FQCN
#               tombstone: t.Optional[{tombstone record}]
#               deprecation: t.Optional[{deprecation record}]
#               redirect: t.Optional[str]
#               redirect_is_symlink: t.Optional[bool]
CollectionRoutingT = t.Mapping[str, t.Mapping[str, t.Mapping[str, t.Any]]]


COLLECTIONS_WITH_FLATMAPPING = (
    'community.general',
    'community.network',
)


def calculate_plugin_fqcns(collection_name: str, src_basename: str,
                           dst_basename: str, rel_path: str) -> t.Tuple[str, str]:
    """
    Calculate source and dest FQCNs for plugins which have been renamed via symlink.

    :collection_name: FQCN of the collection
    :src_basename: Alias name for the plugin
    :dst_basename: Canonical name of the plugin
    :rel_path: Filesystem path to the directory that the plugin lives in relative to the plugin
        type directory.
    :returns: Two-tuple of fqcn of the alias name of the plugin and fqcn of the canonical
        plugin name.
    """
    src_components = os.path.normpath(
        os.path.join(rel_path, src_basename)).split(os.sep)
    dest_components = os.path.normpath(
        os.path.join(rel_path, dst_basename)).split(os.sep)

    # Compose source FQCN
    src_name = '.'.join(src_components)
    src_fqcn = f'{collection_name}.{src_name}'

    # Compose destination FQCN
    dst_name = '.'.join(dest_components)
    dst_fqcn = f'{collection_name}.{dst_name}'

    return src_fqcn, dst_fqcn


def find_symlink_redirects(collection_name: str,
                           plugin_type: str,
                           directory_path: str
                           ) -> t.Dict[str, str]:
    """
    Finds plugin redirects that are defined by symlinks.

    :collection_name: FQCN of the collection we're searching within
    :plugin_type: Type of plugins that we're scanning
    :directory_path: Full path to the directory we're scanning.
    :returns: Dict mapping fqcn of the alias names to fqcn of the canonical plugin names.
    """
    plugin_type_routing = dict()
    if os.path.isdir(directory_path):
        for path, dummy, files in os.walk(directory_path):
            rel_path = os.path.relpath(path, directory_path)
            for filename in files:
                src_basename, ext = os.path.splitext(filename)
                if ext != '.py' and not (plugin_type == 'module' and ext == '.ps1'):
                    continue

                file_path = os.path.join(path, filename)
                if not os.path.islink(file_path):
                    continue

                dst_basename = os.path.splitext(os.readlink(file_path))[0]

                src_fqcn, dst_fqcn = calculate_plugin_fqcns(collection_name, src_basename,
                                                            dst_basename, rel_path)

                plugin_type_routing[src_fqcn] = dst_fqcn

    return plugin_type_routing


def process_dates(plugin_record: t.Dict[str, t.Any]) -> t.Dict[str, t.Any]:
    for tlkey in ('tombstone', 'deprecation'):
        if tlkey in plugin_record and 'removal_date' in plugin_record[tlkey]:
            date = plugin_record[tlkey]['removal_date']
            if isinstance(date, datetime.datetime):
                date = date.date()
            if isinstance(date, datetime.date):
                date = date.isoformat()
            plugin_record[tlkey]['removal_date'] = date
    return plugin_record


def find_flatmapping_short_long_maps(plugin_routing_type: t.Dict[str, t.Dict[str, t.Any]],
                                     ) -> t.Tuple[t.Mapping[str, str],
                                                  t.Mapping[str, t.Tuple[str, bool]]]:
    """
    Collect all short and long names, and mappings between them.

    Short names are FQCN like community.general.rax_facts, and long names are FQCN like
    community.general.cloud.rackspace.rax_facts.

    Returns two tuples. The first element maps short names to long names. The second element
    maps long names to pairs (short name, is symbolic link).
    """
    shortname_to_longname: t.Dict[str, str] = {}
    longname_to_shortname: t.Dict[str, t.Tuple[str, bool]] = {}
    for plugin_name, routing_data in plugin_routing_type.items():
        coll_ns, coll_name, plug_name = get_fqcn_parts(plugin_name)
        if 'tombstone' not in routing_data and 'redirect' in routing_data:
            redirect = routing_data['redirect']
            redir_coll_ns, redir_coll_name, redir_plug_name = get_fqcn_parts(redirect)
            if coll_ns == redir_coll_ns and coll_name == redir_coll_name:
                if '.' not in plug_name and redir_plug_name.endswith(f'.{plug_name}'):
                    is_symlink = routing_data.get('redirect_is_symlink') or False
                    shortname_to_longname[plugin_name] = redirect
                    longname_to_shortname[redirect] = (plugin_name, is_symlink)
                elif '.' in plug_name:
                    # Sometimes plugins/modules/foo_facts.py could be a link to
                    # plugins/modules/subdir1/subdir2/foo_info.py, and
                    # plugins/modules/subdir1/subdir2/foo_facts.py also links to the same
                    # _info module. In that case, artificially construct the shortname
                    # <-> longname mapping
                    dummy, short_name = plug_name.rsplit('.', 1)
                    short_fqcn = f'{coll_ns}.{coll_name}.{short_name}'
                    if plugin_routing_type.get(short_fqcn, {}).get('redirect') == redirect:
                        shortname_to_longname[short_fqcn] = plugin_name
                        longname_to_shortname[plugin_name] = (short_fqcn, False)
    return shortname_to_longname, longname_to_shortname


def remove_flatmapping_artifacts(plugin_routing: t.Dict[str, t.Dict[str, t.Dict[str, t.Any]]]
                                 ) -> None:
    """
    For collections which use flatmapping (like community.general and community.network),
    there will be several redirects which look confusing, like the community.general.rax_facts
    module redirects to community.general.cloud.rackspace.rax_facts, which in turn redirects to
    community.general.cloud.rackspace.rax_info. Such redirects are condensed by this function
    into one redirect from community.general.rax_facts to community.general.rax_info.
    """
    for plugin_type, plugin_routing_type in plugin_routing.items():
        # First collect all short and long names.
        shortname_to_longname, longname_to_shortname = find_flatmapping_short_long_maps(
            plugin_routing_type)
        # Now shorten redirects
        for plugin_name, routing_data in list(plugin_routing_type.items()):
            if 'redirect' in routing_data:
                redirect = routing_data['redirect']
                if shortname_to_longname.get(plugin_name) == redirect:
                    routing_data.pop('redirect')
                    routing_data.pop('redirect_is_symlink', None)
                    if 'deprecation' not in routing_data and 'tombstone' not in routing_data:
                        plugin_routing_type.pop(plugin_name, None)
                elif redirect in longname_to_shortname:
                    routing_data['redirect'], is_symlink = longname_to_shortname[redirect]
                    if routing_data.get('redirect_is_symlink') and not is_symlink:
                        routing_data.pop('redirect_is_symlink')
                if plugin_name in longname_to_shortname:
                    if 'tombstone' not in routing_data and 'deprecation' not in routing_data:
                        plugin_routing_type.pop(plugin_name, None)


def calculate_meta_runtime(collection_name, collection_metadata):
    if collection_name == 'ansible.builtin':
        meta_runtime_path = os.path.join(
            collection_metadata.path, 'config', 'ansible_builtin_runtime.yml')
    else:
        meta_runtime_path = os.path.join(collection_metadata.path, 'meta', 'runtime.yml')

    if os.path.exists(meta_runtime_path):
        meta_runtime = yaml.load_yaml_file(meta_runtime_path)
    else:
        meta_runtime = {}

    return meta_runtime


async def load_collection_routing(collection_name: str,
                                  collection_metadata: AnsibleCollectionMetadata
                                  ) -> t.Dict[str, t.Dict[str, t.Dict[str, t.Any]]]:
    """
    Load plugin routing for a collection.
    """
    meta_runtime = calculate_meta_runtime(collection_name, collection_metadata)
    plugin_routing_out: t.Dict[str, t.Dict[str, t.Dict[str, t.Any]]] = {}
    plugin_routing_in = meta_runtime.get('plugin_routing') or {}
    for plugin_type in DOCUMENTABLE_PLUGINS:
        plugin_type_id = 'modules' if plugin_type == 'module' else plugin_type
        plugin_type_routing = plugin_routing_in.get(plugin_type_id) or {}
        plugin_routing_out[plugin_type] = {
            f'{collection_name}.{plugin_name}': process_dates(plugin_record)
            for plugin_name, plugin_record in plugin_type_routing.items()
        }

    if collection_name == 'ansible.builtin':
        # ansible-base/-core has a special directory structure we currently do not want
        # (or need) to handle
        return plugin_routing_out

    for plugin_type in DOCUMENTABLE_PLUGINS:
        directory_name = 'modules' if plugin_type == 'module' else plugin_type
        directory_path = os.path.join(collection_metadata.path, 'plugins', directory_name)
        plugin_type_routing = plugin_routing_out[plugin_type]

        symlink_redirects = find_symlink_redirects(collection_name, plugin_type, directory_path)
        for redirect_name, redirect_dst in symlink_redirects.items():
            if redirect_name not in plugin_type_routing:
                plugin_type_routing[redirect_name] = {}
            if 'redirect' not in plugin_type_routing[redirect_name]:
                plugin_type_routing[redirect_name]['redirect'] = redirect_dst
            if plugin_type_routing[redirect_name]['redirect'] == redirect_dst:
                plugin_type_routing[redirect_name]['redirect_is_symlink'] = True

    if collection_name in COLLECTIONS_WITH_FLATMAPPING:
        remove_flatmapping_artifacts(plugin_routing_out)

    return plugin_routing_out


async def load_all_collection_routing(collection_metadata: t.Mapping[str,
                                                                     AnsibleCollectionMetadata]
                                      ) -> t.Dict[str, t.Dict[str, t.Dict[str, t.Any]]]:
    # Collection
    lib_ctx = app_context.lib_ctx.get()
    async with asyncio_pool.AioPool(size=lib_ctx.thread_max) as pool:
        requestors = []
        for collection, metadata in collection_metadata.items():
            requestors.append(await pool.spawn(
                load_collection_routing(collection, metadata)))

        responses = await asyncio.gather(*requestors)

    # Merge per-collection routing into one big routing table
    global_plugin_routing: t.Dict[str, t.Dict[str, t.Dict[str, t.Any]]] = {}
    for plugin_type in DOCUMENTABLE_PLUGINS:
        global_plugin_routing[plugin_type] = {}
        for collection_plugin_routing in responses:
            global_plugin_routing[plugin_type].update(collection_plugin_routing[plugin_type])

    return global_plugin_routing


def remove_redirect_duplicates(plugin_info: t.MutableMapping[str, t.MutableMapping[str, t.Any]],
                               collection_routing: CollectionRoutingT) -> None:
    """
    Remove duplicate plugin docs that come from symlinks (or once ansible-docs supports them,
    other plugin routing redirects).
    """
    for plugin_type, plugin_map in plugin_info.items():
        plugin_routing = collection_routing[plugin_type]
        for plugin_name, plugin_record in list(plugin_map.items()):
            # Check redirect
            if plugin_name in plugin_routing and 'redirect' in plugin_routing[plugin_name]:
                destination = plugin_routing[plugin_name]['redirect']
                if destination in plugin_map and destination != plugin_name:
                    # Heuristic: if we have a redirect, and docs for both this plugin and the
                    # redirected one are generated from the same plugin filename, then we can
                    # remove this plugin's docs and generate a redirect stub instead.
                    if compare_all_but(plugin_record['doc'], plugin_map[destination]['doc'],
                                       ['filename']):
                        del plugin_map[plugin_name]


def find_stubs(plugin_info: t.MutableMapping[str, t.MutableMapping[str, t.Any]],
               collection_routing: CollectionRoutingT
               ) -> t.DefaultDict[str, t.DefaultDict[str, t.Dict[str, t.Any]]]:
    """
    Find plugin stubs to write. Returns a nested structure:

        collection:
            plugin_type:
                plugin_short_name:
                    tombstone: t.Optional[{tombstone record}]
                    deprecation: t.Optional[{deprecation record}]
                    redirect: t.Optional[str]
                    redirect_is_symlink: t.Optional[bool]
    """
    stubs_info: t.DefaultDict[str, t.DefaultDict[str, t.Dict[str, t.Any]]] = (
        defaultdict(lambda: defaultdict(dict))
    )
    for plugin_type, plugin_routing in collection_routing.items():
        plugin_info_type = plugin_info.get(plugin_type) or {}
        for plugin_name, plugin_data in plugin_routing.items():
            if 'tombstone' not in plugin_data and 'redirect' not in plugin_data:
                # Ignore pure deprecations
                continue
            if plugin_name not in plugin_info_type:
                coll_ns, coll_name, plug_name = get_fqcn_parts(plugin_name)
                stubs_info[f'{coll_ns}.{coll_name}'][plugin_type][plug_name] = plugin_data

    if 'ansible.builtin' in stubs_info:
        # Remove all redirects from the ansible-base 2.10 -> collections move.
        # We do not want stub pages for all these thousands of plugins/modules.
        ansible_2_10_routing = yaml.load_yaml_bytes(get_antsibull_data('ansible_2_10_routing.yml'))
        for plugin_type, plugins in ansible_2_10_routing['ansible_2_10_routing'].items():
            if plugin_type in stubs_info['ansible.builtin']:
                for plugin_short_name in plugins:
                    stubs_info['ansible.builtin'][plugin_type].pop(plugin_short_name, None)

    return stubs_info



"""
Test the set of Schemas altogether to see that they will parse information correctly.
"""
import glob
import json
import os.path

import pytest

from antsibull.schemas.docs import ansible_doc as ad


# To generate the data in the one_*.json files, run::
#   ansible-doc --json -t $PLUGIN_TYPE $PLUGIN_NAME > \
#       tests/functional/schema/good_data/one_$PLUGINTYPE.json
#
# The matching one_$PLUGINTYPE_results.json file is created by running the schema over the
# input, dumping that to one_$PLUGINTYPE_results.json, and then manually inspecting it to
# make sure it is correct.  Due to the manual inspection, these might be considered better
# at catching regressions than at detecting errors in fresh code.  Some sample python code for
# dumping the file::
#
#   import json
#   from antsibull.schemas.docs.ansible_doc import ConnectionPluginSchema
#   raw = open('one_connection.json').read()
#   normalized = ConnectionPluginSchema.parse_raw(raw)
#   out = json.dumps(normalized.dict(), indent=4, sort_keys=True)
#   open('one_connection_results.json', 'w').write(out)
SINGLE_TESTS = {
    'one_become.json': ad.BecomePluginSchema,
    'one_cache.json': ad.CachePluginSchema,
    'one_callback.json': ad.CallbackPluginSchema,
    'one_cliconf.json': ad.CliConfPluginSchema,
    'one_connection.json': ad.ConnectionPluginSchema,
    'one_httpapi.json': ad.HttpApiPluginSchema,
    'one_inventory.json': ad.InventoryPluginSchema,
    'one_lookup.json': ad.LookupPluginSchema,
    'one_module.json': ad.ModulePluginSchema,
    'one_netconf.json': ad.NetConfPluginSchema,
    'one_shell.json': ad.ShellPluginSchema,
    'one_strategy.json': ad.StrategyPluginSchema,
    'one_vars.json': ad.VarsPluginSchema,
}


@pytest.mark.parametrize('test_file, test_schema', SINGLE_TESTS.items())
def test_one_plugin_of_each_type(test_file, test_schema):
    plugin_type = os.path.splitext(os.path.basename(test_file))[0]
    plugin_type = plugin_type[len('one_'):]

    test_dir = os.path.dirname(__file__)
    result_file = os.path.join(test_dir, 'good_data', 'one_%s_results.json' % plugin_type)
    full_path = os.path.join(test_dir, 'good_data', test_file)

    with open(result_file, 'r') as f:
        results = json.load(f)

    with open(full_path, 'r') as f:
        ansible_doc_output = f.read()

    model = test_schema.parse_raw(ansible_doc_output)

    model_dict = model.dict()
    assert model_dict == results


def test_ssh_connection():
    """Test using the cli field with the ssh connection."""

    # The ssh connection plugin is the only one which can take values directly from
    # dedicated command line arguments in ansible and ansible-playbook.  Make sure that
    # we can deal with its documentation to test that the cli field they use for that works.

    test_file = 'ssh_connection.json'
    test_schema = ad.ConnectionPluginSchema

    test_dir = os.path.dirname(__file__)
    result_file = os.path.join(test_dir, 'good_data', 'ssh_connection_results.json')
    full_path = os.path.join(test_dir, 'good_data', test_file)

    with open(result_file, 'r') as f:
        results = json.load(f)

    with open(full_path, 'r') as f:
        ansible_doc_output = f.read()

    model = test_schema.parse_raw(ansible_doc_output)

    model_dict = model.dict()
    assert model_dict == results



# Copyright: (c) 2019, Ansible Project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
"""
Jinja2 filters for use in Ansible documentation.
"""

import re
from html import escape as html_escape

from jinja2.runtime import Undefined

from ..logging import log


mlog = log.fields(mod=__name__)

# Warning: If you add to this, then you also have to change ansible-doc
# (ansible/cli/__init__.py) in the ansible/ansible repository
_ITALIC = re.compile(r"\bI\(([^)]+)\)")
_BOLD = re.compile(r"\bB\(([^)]+)\)")
_MODULE = re.compile(r"\bM\(([^)]+)\)")
_URL = re.compile(r"\bU\(([^)]+)\)")
_LINK = re.compile(r"\bL\(([^)]+), *([^)]+)\)")
_REF = re.compile(r"\bR\(([^)]+), *([^)]+)\)")
_CONST = re.compile(r"\bC\(([^)]+)\)")
_RULER = re.compile(r"\bHORIZONTALLINE\b")


def html_ify(text):
    ''' convert symbols like I(this is in italics) to valid HTML '''

    flog = mlog.fields(func='html_ify')
    flog.fields(text=text).debug('Enter')
    _counts = {}

    text = html_escape(text)
    text, _counts['italic'] = _ITALIC.subn(r"<em>\1</em>", text)
    text, _counts['bold'] = _BOLD.subn(r"<b>\1</b>", text)
    text, _counts['module'] = _MODULE.subn(r"<span class='module'>\1</span>", text)
    text, _counts['url'] = _URL.subn(r"<a href='\1'>\1</a>", text)
    text, _counts['ref'] = _REF.subn(r"<span class='module'>\1</span>", text)
    text, _counts['link'] = _LINK.subn(r"<a href='\2'>\1</a>", text)
    text, _counts['const'] = _CONST.subn(r"<code>\1</code>", text)
    text, _counts['ruler'] = _RULER.subn(r"<hr/>", text)

    text = text.strip()
    flog.fields(counts=_counts).info('Number of macros converted to html equivalents')
    flog.debug('Leave')
    return text


def documented_type(text):
    ''' Convert any python type to a type for documentation '''

    if isinstance(text, Undefined):
        return '-'
    if text == 'str':
        return 'string'
    if text == 'bool':
        return 'boolean'
    if text == 'int':
        return 'integer'
    if text == 'dict':
        return 'dictionary'
    return text


# The max filter was added in Jinja2-2.10.  Until we can require that version, use this
def do_max(seq):
    return max(seq)


def rst_ify(text):
    ''' convert symbols like I(this is in italics) to valid restructured text '''

    flog = mlog.fields(func='html_ify')
    flog.fields(text=text).debug('Enter')
    _counts = {}

    text, _counts['italic'] = _ITALIC.subn(r"*\1*", text)
    text, _counts['bold'] = _BOLD.subn(r"**\1**", text)
    text, _counts['module'] = _MODULE.subn(r":ref:`\1 <ansible_collections.\1_module>`", text)
    text, _counts['url'] = _LINK.subn(r"`\1 <\2>`_", text)
    text, _counts['ref'] = _URL.subn(r"\1", text)
    text, _counts['link'] = _REF.subn(r":ref:`\1 <\2>`", text)
    text, _counts['const'] = _CONST.subn(r"``\1``", text)
    text, _counts['ruler'] = _RULER.subn(f"\n\n{'-' * 13}\n\n", text)

    flog.fields(counts=_counts).info('Number of macros converted to rst equivalents')
    flog.debug('Leave')
    return text


def rst_fmt(text, fmt):
    ''' helper for Jinja2 to do format strings '''

    return fmt % (text)


def rst_xline(width, char="="):
    ''' return a restructured text line of a given length '''

    return char * width



# coding: utf-8
# Author: Felix Fontein <felix@fontein.de>
# License: GPLv3+
# Copyright: Ansible Project, 2020
'''
Antsibull minimal Sphinx extension which adds some features from the Ansible doc site.
'''

from __future__ import (absolute_import, division, print_function)

__version__ = "0.1.1"
__license__ = "BSD license"
__author__ = "Felix Fontein"
__author_email__ = "felix@fontein.de"


from .assets import setup_assets


def setup(app):
    '''
    Initializer for Sphinx extension API.
    See http://www.sphinx-doc.org/en/stable/extdev/index.html#dev-extensions.
    '''

    # Add assets
    setup_assets(app)

    return dict(
        parallel_read_safe=True,
        parallel_write_safe=True,
        version=__version__,
    )



# coding: utf-8
# Author: Felix Fontein <felix@fontein.de>
# License: GPLv3+
# Copyright: Ansible Project, 2020
'''
Handling assets.
'''

from __future__ import (absolute_import, division, print_function)


import os
import pkgutil

from sphinx.util.osutil import ensuredir


CSS_FILES = [
    'antsibull-minimal.css',
]


def _copy_asset_files(app, exc):  # pylint: disable=unused-argument
    '''
    Copy asset files.
    '''
    # Copy CSS files
    for file in CSS_FILES:
        data = pkgutil.get_data('sphinx_antsibull_ext', file)
        if data is None:
            raise Exception(
                'Internal error: cannot find {0} in sphinx_antsibull_ext package'.format(file))
        ensuredir(os.path.join(app.outdir, '_static'))
        destination = os.path.join(app.outdir, '_static', file)
        with open(destination, 'wb') as f:
            f.write(data)


def setup_assets(app):
    '''
    Setup assets for a Sphinx app object.
    '''
    # Copy assets
    app.connect('build-finished', _copy_asset_files)

    # Add CSS files
    for file in CSS_FILES:
        try:
            app.add_css_file(file)
        except AttributeError:
            # Compat for Sphinx < 1.8
            app.add_stylesheet(file)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['antsibull',
 'antsibull.cli',
 'antsibull.cli.doc_commands',
 'antsibull.data',
 'antsibull.data.docsite',
 'antsibull.docs_parsing',
 'antsibull.jinja2',
 'antsibull.schemas',
 'antsibull.schemas.docs',
 'antsibull.utils',
 'antsibull.vendored',
 'sphinx_antsibull_ext',
 'tests',
 'tests.functional',
 'tests.functional.schema',
 'tests.units']

package_data = \
{'': ['*'],
 'antsibull.data': ['debian/*'],
 'sphinx_antsibull_ext': ['css/*'],
 'tests.functional.schema': ['good_data/*']}

install_requires = \
['PyYAML',
 'aiofiles',
 'aiohttp>=3.0.0',
 'ansible-pygments',
 'antsibull-changelog>=0.7.0',
 'asyncio-pool',
 'docutils',
 'jinja2',
 'packaging>=20.0',
 'perky',
 'pydantic',
 'rstcheck>=3,<4',
 'semantic_version',
 'sh',
 'sphinx',
 'twiggy>=0.5.0']

extras_require = \
{':python_version < "3.8"': ['importlib-metadata'],
 ':python_version >= "3.6" and python_version < "3.7"': ['aiocontextvars']}

entry_points = \
{'console_scripts': ['antsibull-build = antsibull.cli.antsibull_build:main',
                     'antsibull-docs = antsibull.cli.antsibull_docs:main',
                     'antsibull-lint = antsibull.cli.antsibull_lint:main']}

setup_kwargs = {
    'name': 'antsibull',
    'version': '0.36.0',
    'description': 'Tools for building the Ansible Distribution',
    'long_description': "# antsibull -- Ansible Build Scripts\n[![Python linting badge](https://github.com/ansible-community/antsibull/workflows/Python%20linting/badge.svg?event=push&branch=main)](https://github.com/ansible-community/antsibull/actions?query=workflow%3A%22Python+linting%22+branch%3Amain)\n[![Python testing badge](https://github.com/ansible-community/antsibull/workflows/Python%20testing/badge.svg?event=push&branch=main)](https://github.com/ansible-community/antsibull/actions?query=workflow%3A%22Python+testing%22+branch%3Amain)\n[![Build CSS testing badge](https://github.com/ansible-community/antsibull/workflows/Build%20CSS/badge.svg?event=push&branch=main)](https://github.com/ansible-community/antsibull/actions?query=workflow%3A%22Build+CSS%22+branch%3Amain)\n[![dumb PyPI on GH pages badge](https://github.com/ansible-community/antsibull/workflows/👷%20dumb%20PyPI%20on%20GH%20pages/badge.svg?event=push&branch=main)](https://github.com/ansible-community/antsibull/actions?query=workflow%3A%22👷+dumb+PyPI+on+GH+pages%22+branch%3Amain)\n[![Codecov badge](https://img.shields.io/codecov/c/github/ansible-community/antsibull)](https://codecov.io/gh/ansible-community/antsibull)\n\nTooling for building various things related to Ansible\n\nScripts that are here:\n\n* antsibull-build - Builds Ansible-2.10+ from component collections ([docs](docs/build-ansible.rst))\n* antsibull-docs - Extracts documentation from ansible plugins\n* antsibull-lint - Right now only validates ``changelogs/changelog.yaml`` files ([docs](docs/changelog.yaml-format.md))\n\nThis also includes a [Sphinx extension](https://www.sphinx-doc.org/en/master/) `sphinx_antsibull_ext` which provides a minimal CSS file to render the output of `antsibull-docs` correctly.\n\nA related project is [antsibull-changelog](https://pypi.org/project/antsibull-changelog/), which is in its [own repository](https://github.com/ansible-community/antsibull-changelog/).\n\nScripts are created by poetry at build time.  So if you want to run from\na checkout, you'll have to run them under poetry::\n\n    python3 -m pip install poetry\n    poetry install  # Installs dependencies into a virtualenv\n    poetry run antsibull-build --help\n\n.. note:: When installing a package published by poetry, it is best to use\n    pip >= 19.0.  Installing with pip-18.1 and below could create scripts which\n    use pkg_resources which can slow down startup time (in some environments by\n    quite a large amount).\n\nUnless otherwise noted in the code, it is licensed under the terms of the GNU\nGeneral Public License v3 or, at your option, later.\n\nantsibull is covered by the [Ansible Code of Conduct](https://docs.ansible.com/ansible/latest/community/code_of_conduct.html).\n\n## Using the Sphinx extension\n\nInclude it in your Sphinx configuration ``conf.py``::\n\n```\n# Add it to 'extensions':\nextensions = ['sphinx.ext.autodoc', 'sphinx.ext.intersphinx', 'notfound.extension', 'sphinx_antsibull_ext']\n```\n\n## Updating the CSS file for the Sphinx extension\n\nThe CSS file [sphinx_antsibull_ext/antsibull-minimal.css](https://github.com/ansible-community/antsibull/blob/main/sphinx_antsibull_ext/antsibull-minimal.css) is built from [sphinx_antsibull_ext/css/antsibull-minimal.scss](https://github.com/ansible-community/antsibull/blob/main/sphinx_antsibull_ext/src/antsibull-minimal.scss) using [SASS](https://sass-lang.com/) and [postcss](https://postcss.org/) using [autoprefixer](https://github.com/postcss/autoprefixer) and [cssnano](https://cssnano.co/).\n\nUse the script `build.sh` in `sphinx_antsibull_ext/css/` to build the `.css` file from the `.scss` file:\n\n```\ncd sphinx_antsibull_ext/css/\n./build-css.sh\n```\n\nFor this to work, you need to make sure that `sassc` and `postcss` are on your path and that the autoprefixer and nanocss modules are installed:\n\n```\n# Debian:\napt-get install sassc\n\n# PostCSS, autoprefixer and cssnano require nodejs/npm:\nnpm install -g autoprefixer cssnano postcss postcss-cli\n```\n\n## Creating a new release:\n\nIf you want to create a new release::\n\n    poetry build\n    poetry publish  # Uploads to pypi.  Be sure you really want to do this\n\n    git tag $VERSION_NUMBER\n    git push --tags\n    vim pyproject.toml    # Bump the version number\n    git commit -m 'Update the version number for the next release' pyproject.toml\n    git push\n",
    'author': 'Toshio Kuratomi',
    'author_email': 'a.badger@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/ansible-community/antsibull',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.6.0,<4.0.0',
}


setup(**setup_kwargs)



import os

import twiggy.levels

import antsibull.logging as al


def test_library_settings(capsys):
    """Importing without calling initialize_app_logging leaves logging disabled."""
    assert al.log.min_level == twiggy.levels.DISABLED

    al.log.error('test')
    captured = capsys.readouterr()
    assert captured.err == ''
    assert captured.out == ''


def test_initialize_app_logging(capsys):
    """Callind initialize_app_logging outputs WARNING and above to stderr."""
    al.initialize_app_logging()

    al.log.error('test')
    al.log.warning('second test')
    al.log.debug('third test')
    captured = capsys.readouterr()
    assert captured.err == 'ERROR:antsibull|test\nWARNING:antsibull|second test\n'
    assert captured.out == ''


def test_initialize_app_logging_debug(capsys):
    os.environ['ANTSIBULL_EARLY_DEBUG'] = '1'
    al.initialize_app_logging()

    al.log.error('test')
    al.log.warning('second test')
    al.log.debug('third test')
    captured = capsys.readouterr()
    assert captured.err == 'ERROR:antsibull|test\nWARNING:antsibull|second test\nDEBUG:antsibull|third test\n'
    assert captured.out == ''



import asyncio
import json
from unittest.mock import patch

import pytest
from asynctest import CoroutineMock
from aiohttp_utils import CaseControlledTestServer, http_redirect
from certificate_utils import ssl_certificate

from antsibull.galaxy import GalaxyClient


SAMPLE_VERSIONS = {
    "count": 1,
    "next": None,
    "previous": None,
    "results": [
        {
            "version": "0.1.1",
            "href": "https://galaxy.ansible.com/api/v2/collections/community/general/versions/0.1.1/"
        }
    ]
}


@pytest.mark.asyncio
async def test_get_collection_version_info(http_redirect, ssl_certificate):
    async with CaseControlledTestServer(ssl=ssl_certificate.server_context()) as server:
        http_redirect.add_server('galaxy.ansible.com', 443, server.port)
        gc = GalaxyClient(aio_session=http_redirect.session)
        task = asyncio.ensure_future(gc.get_versions('community.general'))

        request = await server.receive_request(timeout=5)
        assert request.path_qs == '/api/v2/collections/community/general/versions/?format=json&page_size=100'

        server.send_response(request, text=json.dumps(SAMPLE_VERSIONS),
                             headers={'Content-Type': 'application/json'})
        assert await task == ['0.1.1']



# Author: Toshio Kuratomi <tkuratom@redhat.com>
# License: GPLv3+
# Copyright: Ansible Project, 2020
"""Augment data from plugin documenation with additional values."""

import typing as t


def add_full_key(options_data: t.Mapping[str, t.Any], suboption_entry: str,
                 _full_key: t.Optional[list] = None) -> None:
    """
    Add information on the strucfture of a dict value in options or returns.

    suboptions and contains are used for nested information (an option taking a dict which has
    a deeply nested structure, for instance.)  They describe each entry into the dict.  When
    constructing documentation which uses that, it can be useful to know the hierarchy leads to
    that entry (for instance, to make a target for an html href).  This function adds that
    information to a ``full_key`` field on the suboptions' entry.

    :arg options_data: The documentation data which is going to be analyzed and updated.
    :arg suboption_entry: The name of the suboptions key in the data.  For options data, this is
        ``suboptions``.  For returndocs, it is ``contains``.
    :kwarg _full_key: This is a recursive function.  After we pass the first level of nesting,
        ``_full_key`` is set to record the names of the upper levels of the hierarchy.

    .. warning:: This function operates by side-effect.  The options_data dictionay is modified
        directly.
    """
    if _full_key is None:
        _full_key = []

    for (key, entry) in options_data.items():
        # Make sure that "full key" is contained
        full_key_k = _full_key + [key]
        entry['full_key'] = full_key_k

        # Process suboptions
        suboptions = entry.get(suboption_entry)
        if suboptions:
            add_full_key(suboptions, suboption_entry=suboption_entry, _full_key=full_key_k)


def augment_docs(plugin_info: t.MutableMapping[str, t.MutableMapping[str, t.Any]]) -> None:
    """
    Add additional data to the data extracted from the plugins.

    The additional data is calculated from the existing data and then added to the data.
    Current Augmentations:

    * ``full_key`` allows displaying nested suboptions and return dicts.

    :arg plugin_info: The plugin_info that will be analyzed and augmented.

    .. warning:: This function operates by side-effect.  The plugin_info dictionay is modified
        directly.
    """
    for plugin_type, plugin_map in plugin_info.items():
        for plugin_name, plugin_record in plugin_map.items():
            add_full_key(plugin_record['return'], 'contains')
            add_full_key(plugin_record['doc']['options'], 'suboptions')



import pytest

from antsibull.jinja2.filters import rst_ify


RST_IFY_DATA = {
    # No substitutions
    'no-op': 'no-op',
    'no-op Z(test)': 'no-op Z(test)',
    # Simple cases of all substitutions
    'I(italic)': '*italic*',
    'B(bold)': '**bold**',
    'M(ansible.builtin.yum)': ':ref:`ansible.builtin.yum'
    ' <ansible_collections.ansible.builtin.yum_module>`',
    'U(https://docs.ansible.com)': 'https://docs.ansible.com',
    'L(the user guide,https://docs.ansible.com/user-guide.html)': '`the user guide'
    ' <https://docs.ansible.com/user-guide.html>`_',
    'R(the user guide,user-guide)': ':ref:`the user guide <user-guide>`',
    'C(/usr/bin/file)': '``/usr/bin/file``',
    'HORIZONTALLINE': '\n\n{0}\n\n'.format('-' * 13),
    # Multiple substitutions
    'The M(ansible.builtin.yum) module B(MUST) be given the C(package) parameter.  See the R(looping docs,using-loops) for more info':
    'The :ref:`ansible.builtin.yum <ansible_collections.ansible.builtin.yum_module>` module **MUST** be given the ``package`` parameter.  See the :ref:`looping docs <using-loops>` for more info',
    # Problem cases
    'IBM(International Business Machines)': 'IBM(International Business Machines)',
    'L(the user guide, https://docs.ansible.com/)': '`the user guide <https://docs.ansible.com/>`_',
    'R(the user guide, user-guide)': ':ref:`the user guide <user-guide>`',
}


@pytest.mark.parametrize('text, expected', RST_IFY_DATA.items())
def test_rst_ify(text, expected):
    assert rst_ify(text) == expected



# coding: utf-8
# Author: Toshio Kuratomi <tkuratom@redhat.com>
# License: GPLv3+
# Copyright: Ansible Project, 2020
"""Functions to work with Galaxy."""

import os.path
import shutil
import typing as t
from urllib.parse import urljoin

import semantic_version as semver

from . import app_context
from .hashing import verify_hash
from .utils.http import retry_get

# The type checker can handle finding aiohttp.client but flake8 cannot :-(
if t.TYPE_CHECKING:
    import aiohttp.client


#: URL to galaxy. (Get the default from the context default)
_GALAXY_SERVER_URL = str(app_context.AppContext().galaxy_url)


class NoSuchCollection(Exception):
    """Collection name does not map to a collection on Galaxy."""


class NoSuchVersion(Exception):
    """Version does not match with any versions of a collection on Galaxy."""


class DownloadFailure(Exception):
    """Failure downloading a collection from Galaxy."""


class DownloadResults(t.NamedTuple):
    """Results of downloading a collection."""

    #: :obj:`semantic_version.Version` of the exact version of the collection that was downloaded.
    version: semver.Version
    #: Location on the filesystem of the downloaded collection.
    download_path: str


class GalaxyClient:
    """Class for querying the Galaxy REST API."""

    def __init__(self, aio_session: 'aiohttp.client.ClientSession',
                 galaxy_server: str = _GALAXY_SERVER_URL) -> None:
        """
        Create a GalaxyClient object to query the Galaxy Server.

        :arg aio_session: :obj:`aiohttp.ClientSession` with which to perform all
            requests to galaxy.
        :kwarg galaxy_server: URL to the galaxy server.
        """
        self.galaxy_server = galaxy_server
        self.aio_session = aio_session
        self.params = {'format': 'json'}

    async def _get_galaxy_versions(self, versions_url: str) -> t.List[str]:
        """
        Retrieve the complete list of versions for a collection from a galaxy endpoint.

        This internal function retrieves versions for collections from a Galaxy endpoint.  If the
        information is paged, it continues to retrieve linked pages until all of the information has
        been returned.

        :arg version_url: url to the page to retrieve.
        :returns: List of the all the versions of the collection.
        """
        params = self.params.copy()
        params['page_size'] = '100'
        async with retry_get(self.aio_session, versions_url, params=params,
                             acceptable_error_codes=[404]) as response:
            if response.status == 404:
                raise NoSuchCollection(f'No collection found at: {versions_url}')
            collection_info = await response.json()

        versions = []
        for version_record in collection_info['results']:
            versions.append(version_record['version'])

        if collection_info['next']:
            versions.extend(await self._get_galaxy_versions(collection_info['next']))

        return versions

    async def get_versions(self, collection: str) -> t.List[str]:
        """
        Retrieve all versions of a collection on Galaxy.

        :arg collection: Name of the collection to get version info for.
        :returns: List of all the versions of this collection on galaxy.
        """
        collection = collection.replace('.', '/')
        galaxy_url = urljoin(self.galaxy_server, f'api/v2/collections/{collection}/versions/')
        retval = await self._get_galaxy_versions(galaxy_url)
        return retval

    async def get_info(self, collection: str) -> t.Dict[str, t.Any]:
        """
        Retrieve information about the collection on Galaxy.

        :arg collection: Namespace.collection to retrieve information about.
        :returns: Dictionary of information about the collection.

        Please see the Galaxy REST API documentation for information on the structure of the
        returned data.

        .. seealso::
            An example return value from the
            `Galaxy REST API <https://galaxy.ansible.com/api/v2/collections/community/general/>`_
        """
        collection = collection.replace('.', '/')
        galaxy_url = urljoin(self.galaxy_server, f'api/v2/collections/{collection}/')

        async with retry_get(self.aio_session, galaxy_url, params=self.params,
                             acceptable_error_codes=[404]) as response:
            if response.status == 404:
                raise NoSuchCollection(f'No collection found at: {galaxy_url}')
            collection_info = await response.json()

        return collection_info

    async def get_release_info(self, collection: str,
                               version: t.Union[str, semver.Version]) -> t.Dict[str, t.Any]:
        """
        Retrive information about a specific version of a collection.

        :arg collection: Namespace.collection string naming the collection.
        :arg version: Version of the collection.
        :returns: Dictionary of information about the release.

        Please see the Galaxy REST API documentation for information on the structure of the
        returned data.

        .. seealso::
            An example return value from the
            `Galaxy REST API
            <https://galaxy.ansible.com/api/v2/collections/community/general/versions/0.1.1>`_
        """
        collection = collection.replace('.', '/')
        galaxy_url = urljoin(self.galaxy_server,
                             f'api/v2/collections/{collection}/versions/{version}/')

        async with retry_get(self.aio_session, galaxy_url, params=self.params,
                             acceptable_error_codes=[404]) as response:
            if response.status == 404:
                raise NoSuchCollection(f'No collection found at: {galaxy_url}')
            collection_info = await response.json()

        return collection_info

    async def get_latest_matching_version(self, collection: str,
                                          version_spec: str,
                                          pre: bool = False) -> semver.Version:
        """
        Get the latest version of a collection that matches a specification.

        :arg collection: Namespace.collection identifying a collection.
        :arg version_spec: String specifying the allowable versions.
        :kwarg pre: If True, allow prereleases (versions which have the form X.Y.Z.SOMETHING).
            This is **not** for excluding 0.Y.Z versions.  The default is False.
        :returns: :obj:`semantic_version.Version` of the latest collection version that satisfied
            the specification.

        .. seealso:: For the format of the version_spec, see the documentation
            of :obj:`semantic_version.SimpleSpec`
        """
        versions = await self.get_versions(collection)
        versions = [semver.Version(v) for v in versions]
        versions.sort(reverse=True)

        spec = semver.SimpleSpec(version_spec)
        for version in (v for v in versions if v in spec):
            # If we're excluding prereleases and this is a prerelease, then skip it.
            if not pre and version.prerelease:
                continue
            return version

        # No matching versions were found
        raise NoSuchVersion(f'{version_spec} did not match with any version of {collection}.')


class CollectionDownloader(GalaxyClient):
    """Manage downloading collections from Galaxy."""

    def __init__(self, aio_session: 'aiohttp.client.ClientSession',
                 download_dir: str,
                 galaxy_server: str = _GALAXY_SERVER_URL,
                 collection_cache: t.Optional[str] = None) -> None:
        """
        Create an object to download collections from galaxy.

        :arg aio_session: :obj:`aiohttp.ClientSession` with which to perform all
            requests to galaxy.
        :arg download_dir: Directory to download into.
        :kwarg galaxy_server: URL to the galaxy server.
        :kwarg collection_cache: If given, a path to a directory containing collection tarballs.
            These tarballs will be used instead of downloading new tarballs provided that the
            versions match the criteria (latest compatible version known to galaxy).
        """
        super().__init__(aio_session, galaxy_server)
        self.download_dir = download_dir
        # TODO: PY3.8: self.collection_cache: t.Final[t.Optional[str]] = collection_cache
        self.collection_cache = collection_cache

    async def download(self, collection: str, version: t.Union[str, semver.Version], ) -> str:
        """
        Download a collection.

        Downloads the collection at the specified version.

        :arg collection: Namespace.collection identifying the collection.
        :arg version: Version of the collection to download.
        :returns: The full path to the downloaded collection.
        """
        collection = collection.replace('.', '/')
        release_info = await self.get_release_info(collection, version)
        release_url = release_info['download_url']

        download_filename = os.path.join(self.download_dir, release_info['artifact']['filename'])
        sha256sum = release_info['artifact']['sha256']

        if self.collection_cache:
            if release_info['artifact']['filename'] in os.listdir(self.collection_cache):
                # TODO: PY3.8: We can use t.Final in __init__ instead of cast here.
                cached_copy = os.path.join(t.cast(str, self.collection_cache),
                                           release_info['artifact']['filename'])
                if await verify_hash(cached_copy, sha256sum):
                    shutil.copyfile(cached_copy, download_filename)
                return download_filename

        async with retry_get(self.aio_session, release_url,
                             acceptable_error_codes=[404]) as response:
            if response.status == 404:
                raise NoSuchCollection(f'No collection found at: {release_url}')

            with open(download_filename, 'wb') as f:
                lib_ctx = app_context.lib_ctx.get()
                # TODO: PY3.8: while chunk := await response.content.read(lib_ctx.chunksize):
                chunk = await response.content.read(lib_ctx.chunksize)
                while chunk:
                    f.write(chunk)
                    chunk = await response.content.read(lib_ctx.chunksize)

        # Verify the download
        if not await verify_hash(download_filename, sha256sum):
            raise DownloadFailure(f'{release_url} failed to download correctly.'
                                  f' Expected checksum: {sha256sum}')

        # Copy downloaded collection into cache
        if self.collection_cache:
            # TODO: PY3.8: We can use t.Final in __init__ instead of cast here.
            cached_copy = os.path.join(t.cast(str, self.collection_cache),
                                       release_info['artifact']['filename'])
            shutil.copyfile(download_filename, cached_copy)

        return download_filename

    async def download_latest_matching(self, collection: str,
                                       version_spec: str) -> DownloadResults:
        """
        Download the latest version of a collection that matches a specification.

        :arg collection: Namespace.collection identifying a collection.
        :arg version_spec: String specifying the allowable versions.
        :returns: :obj:`DownloadResults` with version and download path for the collection we
            downloaded.

        .. seealso:: For the format of the version_spec, see the documentation
            of :obj:`semantic_version.SimpleSpec`
        """
        version = await self.get_latest_matching_version(collection, version_spec)
        download_path = await self.download(collection, version)
        return DownloadResults(version=version, download_path=download_path)



# Copyright: (c) 2014, James Tanner <tanner.jc@gmail.com>
# Copyright: (c) 2018, Ansible Project
# Copyright: (c) 2020, Felix Fontein
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

# Parts taken from Ansible's ansible-doc sources

import argparse
import json
import os
import sys

import yaml

import ansible.plugins.loader as plugin_loader

from ansible import constants as C
from ansible import release as ansible_release
from ansible.cli import doc
from ansible.cli.arguments import option_helpers as opt_help
from ansible.collections.list import list_collection_dirs
from ansible.module_utils._text import to_native
from ansible.module_utils.common.json import AnsibleJSONEncoder
from ansible.plugins.loader import action_loader, fragment_loader
from ansible.utils.collection_loader import AnsibleCollectionConfig
from ansible.utils.plugin_docs import get_docstring


def load_plugin(loader, plugin_type, plugin):
    result = {}
    try:
        plugin_context = loader.find_plugin_with_context(
            plugin, mod_type='.py', ignore_deprecated=True, check_aliases=True)
        if not plugin_context.resolved:
            result['error'] = 'Cannot find plugin'
            return result
        plugin_name = plugin_context.plugin_resolved_name
        filename = plugin_context.plugin_resolved_path
        collection_name = plugin_context.plugin_resolved_collection

        result.update({
            'plugin_name': plugin_name,
            'filename': filename,
            'collection_name': collection_name,
        })

        documentation, plainexamples, returndocs, metadata = get_docstring(
            filename, fragment_loader, verbose=False,
            collection_name=collection_name, is_module=(plugin_type == 'module'))

        if documentation is None:
            result['error'] = 'No valid documentation found'
            return result

        documentation['filename'] = filename
        documentation['collection'] = collection_name

        if plugin_type == 'module':
            # is there corresponding action plugin?
            if plugin in action_loader:
                documentation['has_action'] = True
            else:
                documentation['has_action'] = False

        ansible_doc = {
            'doc': documentation,
            'examples': plainexamples,
            'return': returndocs,
            'metadata': metadata,
        }

        try:
            # If this fails, the documentation cannot be seralized as JSON
            json.dumps(ansible_doc, cls=AnsibleJSONEncoder)
            # Store result. This is guaranteed to be serializable
            result['ansible-doc'] = ansible_doc
        except Exception as e:
            result['error'] = (
                'Cannot serialize documentation as JSON: %s' % to_native(e)
            )
    except Exception as e:
        result['error'] = (
            'Missing documentation or could not parse documentation: %s' % to_native(e)
        )

    return result


def ansible_doc_coll_filter(coll_filter):
    return coll_filter[0] if coll_filter and len(coll_filter) == 1 else None


def match_filter(name, coll_filter):
    if coll_filter is None or name in coll_filter:
        return True
    if '.' not in name and 'ansible.builtin' in coll_filter:
        return True
    for filter in coll_filter:
        if name.startswith(filter + '.'):
            return True
    return False


def load_all_plugins(plugin_type, basedir, coll_filter):
    loader = getattr(plugin_loader, '%s_loader' % plugin_type)

    if basedir:
        loader.add_directory(basedir, with_subdir=True)

    loader._paths = None  # reset so we can use subdirs below

    plugin_list = set()

    if match_filter('ansible.builtin', coll_filter):
        paths = loader._get_paths_with_context()
        for path_context in paths:
            plugin_list.update(
                doc.DocCLI.find_plugins(path_context.path, path_context.internal, plugin_type))

    doc.add_collection_plugins(
        plugin_list, plugin_type, coll_filter=ansible_doc_coll_filter(coll_filter))

    result = {}
    for plugin in plugin_list:
        if match_filter(plugin, coll_filter):
            result[plugin] = load_plugin(loader, plugin_type, plugin)

    return result


def load_collection_meta_manifest(b_manifest_path):
    with open(b_manifest_path, 'rb') as f:
        meta = json.load(f)
    return {
        'namespace': meta['collection_info']['namespace'],
        'name': meta['collection_info']['name'],
        'version': meta['collection_info']['version'],
    }


def load_collection_meta_galaxy(b_galaxy_path):
    with open(b_galaxy_path, 'rb') as f:
        meta = yaml.safe_load(f)
    return {
        'namespace': meta['namespace'],
        'name': meta['name'],
        'version': meta.get('version'),
    }


def load_collection_meta(b_path):
    b_manifest_path = os.path.join(b_path, b'MANIFEST.json')
    if os.path.exists(b_manifest_path):
        data = load_collection_meta_manifest(b_manifest_path)
    else:
        data = {}
        b_galaxy_path = os.path.join(b_path, b'galaxy.yml')
        b_galaxy_alt_path = os.path.join(b_path, b'galaxy.yaml')
        for path in (b_galaxy_path, b_galaxy_alt_path):
            if os.path.exists(path):
                data = load_collection_meta_galaxy(path)
        if not data:
            # Fallback in case no galaxy.yml is around. This happens for collections
            # checked out from source where galaxy.yml is templated on build.
            data = {
                'version': None,
                'namespace': to_native(os.path.basename(os.path.dirname(b_path))),
                'name': to_native(os.path.basename(b_path)),
            }
    data['path'] = to_native(b_path)
    return data


def main(args):
    parser = argparse.ArgumentParser(
        prog=args[0], description='Bulk extraction of Ansible plugin docs.')
    parser.add_argument('args', nargs='*', help='Collection filter', metavar='collection_filter')
    parser.add_argument('--pretty', action='store_true', help='Pretty-print JSON')
    opt_help.add_basedir_options(parser)

    arguments = parser.parse_args(args[1:])

    basedir = arguments.basedir
    coll_filter = arguments.args or None

    if basedir:
        AnsibleCollectionConfig.playbook_paths = basedir

    result = {
        'plugins': {},
        'collections': {},
    }

    # Export plugin docs
    for plugin_type in C.DOCUMENTABLE_PLUGINS:
        result['plugins'][plugin_type] = load_all_plugins(plugin_type, basedir, coll_filter)

    # Export collection data
    b_colldirs = list_collection_dirs(coll_filter=ansible_doc_coll_filter(coll_filter))
    for b_path in b_colldirs:
        meta = load_collection_meta(b_path)
        collection_name = '{0}.{1}'.format(meta['namespace'], meta['name'])
        if match_filter(collection_name, coll_filter):
            result['collections'][collection_name] = meta
    if match_filter('ansible.builtin', coll_filter):
        result['collections']['ansible.builtin'] = {
            'path': os.path.dirname(ansible_release.__file__),
            'version': ansible_release.__version__,
        }

    print(json.dumps(
        result, cls=AnsibleJSONEncoder, sort_keys=True, indent=4 if arguments.pretty else None))


if __name__ == '__main__':
    main(sys.argv)



# coding: utf-8
# Author: Toshio Kuratomi <tkuratom@redhat.com>
# License: GPLv3+
# Copyright: Ansible Project, 2020
"""
Schemas for the plugin DOCUMENTATION data.

This is a highlevel interface.  The hope is that developers can use either this or
antsibull.schemas.docs.ansible_doc to handle all of their validation needs.
"""

from .callback import CallbackDocSchema, CallbackSchema
from .module import ModuleDocSchema, ModuleSchema
from .plugin import (PluginDocSchema, PluginExamplesSchema,
                     PluginMetadataSchema, PluginReturnSchema, PluginSchema)

BecomeSchema = PluginSchema
CacheSchema = PluginSchema
CliConfSchema = PluginSchema
ConnectionSchema = PluginSchema
HttpApiSchema = PluginSchema
InventorySchema = PluginSchema
LookupSchema = PluginSchema
NetConfSchema = PluginSchema
ShellSchema = PluginSchema
StrategySchema = PluginSchema
VarsSchema = PluginSchema


#: The schemas that most plugins use to validate and normalize their documentation.
_PLUGIN_SCHEMA_RECORD = {
    'top': PluginSchema,
    'doc': PluginDocSchema,
    'examples': PluginExamplesSchema,
    'metadata': PluginMetadataSchema,
    'return': PluginReturnSchema,
}


#: Mapping of plugin_types to the schemas which validate and normalize their documentation.
#: The structure of this mapping is a two level nested dict.  The outer key is the plugin_type.
#: The inner keys are the sections of the documentation (doc, example, metadata, return, or top
#: [which combines all of hte above, such as ansible-doc returns]) to validate.
DOCS_SCHEMAS = {
    'become': _PLUGIN_SCHEMA_RECORD,
    'cache': _PLUGIN_SCHEMA_RECORD,
    'callback': {
        'top': CallbackSchema,
        'doc': CallbackDocSchema,
        'examples': PluginExamplesSchema,
        'metadata': PluginMetadataSchema,
        'return': PluginReturnSchema,
    },
    'cliconf': _PLUGIN_SCHEMA_RECORD,
    'connection': _PLUGIN_SCHEMA_RECORD,
    'httpapi': _PLUGIN_SCHEMA_RECORD,
    'inventory': _PLUGIN_SCHEMA_RECORD,
    'lookup': _PLUGIN_SCHEMA_RECORD,
    'module': {
        'top': ModuleSchema,
        'doc': ModuleDocSchema,
        'examples': PluginExamplesSchema,
        'metadata': PluginMetadataSchema,
        'return': PluginReturnSchema,
    },
    'netconf': _PLUGIN_SCHEMA_RECORD,
    'shell': _PLUGIN_SCHEMA_RECORD,
    'strategy': _PLUGIN_SCHEMA_RECORD,
    'vars': _PLUGIN_SCHEMA_RECORD,
}



# coding: utf-8
# Author: Toshio Kuratomi <tkuratom@redhat.com>
# License: GPLv3+
# Copyright: Ansible Project, 2020
"""Functions to handle config files."""

import itertools
import os.path
import typing as t

import perky

from .logging import log
from .schemas.config import ConfigModel


mlog = log.fields(mod=__name__)

#: System config file location.
SYSTEM_CONFIG_FILE = '/etc/antsibull.cfg'

#: Per-user config file location.
USER_CONFIG_FILE = '~/.antsibull.cfg'


def find_config_files(conf_files: t.Iterable[str]) -> t.List[str]:
    """
    Find all config files that exist.

    :arg conf_files: An iterable of config filenames to search for.
    :returns: A List of filenames which actually existed on the system.
    """
    flog = mlog.fields(func='find_config_file')
    flog.fields(conf_files=conf_files).debug('Enter')

    paths = [os.path.abspath(p) for p in conf_files]
    flog.fields(paths=paths).info('Paths to check')

    config_files = []
    for conf_path in paths:
        if os.path.exists(conf_path):
            config_files.append(conf_path)
    flog.fields(paths=config_files).info('Paths found')

    flog.debug('Leave')
    return config_files


def read_config(filename: str) -> ConfigModel:
    """
    Parse a config file and return the data from it.

    :arg filename: The filename of the config file to parse.
    :returns: A ConfigModel model containing the config data.
    """
    flog = mlog.fields(func='read_config')
    flog.debug('Enter')

    filename = os.path.abspath(filename)

    flog.fields(filename=filename).info('loading config file')
    raw_config_data = perky.load(filename)
    flog.debug('Validatinging the config file data')
    # Note: We parse the object but discard the model because we want to validate the config but let
    # the context handle all setting of defaults
    ConfigModel.parse_obj(raw_config_data)

    flog.debug('Leave')
    return raw_config_data


def load_config(conf_files: t.Union[t.Iterable[str], str, None] = None) -> t.Dict:
    """
    Load configuration.

    Load configuration from all found conf files.  The default configuration is loaded
    followed by a system-wide location, user-location, and then any files specified in
    the ``conf_files`` parameter.  Toplevel keys in later files will overwrite earlier
    those same keys in earlier files.

    :arg conf_files: An iterable of conf_files to load configuration information from.
    :returns: A dict containing the configuration.
    """
    flog = mlog.fields(func='load_config')
    flog.debug('Enter')

    if isinstance(conf_files, str):
        conf_files = (conf_files,)
    elif conf_files is None:
        conf_files = ()

    user_config_file = os.path.expanduser(USER_CONFIG_FILE)
    available_files = find_config_files(itertools.chain((SYSTEM_CONFIG_FILE, user_config_file),
                                                        conf_files))

    includes = list(available_files)

    flog.debug('loading config files')
    # Perky has some bugs that prevent this simple way from working:
    # https://github.com/ansible-community/antsibull/pull/118
    # cfg = {'includes': includes}
    # cfg = perky.includes(cfg, recursive=True)

    # Workaround for above bug.  Note that includes specified in the config files will not work
    # but we can just add that as a new feature when perky gets it working.
    cfg = {}
    for filename in includes:
        new_cfg = perky.load(filename)
        cfg.update(new_cfg)

    flog.debug('validating configuration')
    # Note: We parse the object but discard the model because we want to validate the config but let
    # the context handle all setting of defaults
    ConfigModel.parse_obj(cfg)

    flog.fields(config=cfg).debug('Leave')
    return cfg



# coding: utf-8
# Author: Toshio Kuratomi <tkuratom@redhat.com>
# License: GPLv3+
# Copyright: Ansible Project, 2020
"""Entrypoint to the antsibull-build tool."""

import argparse
import os.path
import sys
from typing import List

import twiggy
from packaging.version import Version as PypiVer

from ..logging import log, initialize_app_logging
initialize_app_logging()

# We have to call initialize_app_logging() before these imports so that the log object is configured
# correctly before other antisbull modules make copies of it.
# pylint: disable=wrong-import-position
from .. import app_context  # noqa: E402
from ..args import (  # noqa: E402
    InvalidArgumentError, get_toplevel_parser, normalize_toplevel_options
)
from ..build_collection import build_collection_command  # noqa: E402
from ..build_ansible_commands import (  # noqa: E402
    build_single_command, build_multiple_command, rebuild_single_command,
)
from ..build_changelog import build_changelog  # noqa: E402
from ..config import load_config  # noqa: E402
from ..new_ansible import new_ansible_command  # noqa: E402
# pylint: enable=wrong-import-position


mlog = log.fields(mod=__name__)

DEFAULT_FILE_BASE = 'ansible'
DEFAULT_PIECES_FILE = f'{DEFAULT_FILE_BASE}.in'

ARGS_MAP = {'new-ansible': new_ansible_command,
            'single': build_single_command,
            'multiple': build_multiple_command,
            'collection': build_collection_command,
            'changelog': build_changelog,
            'rebuild-single': rebuild_single_command,
            # Old names, deprecated
            'new-acd': new_ansible_command,
            'build-single': build_single_command,
            'build-multiple': build_multiple_command,
            'build-collection': build_collection_command,
            }


def _normalize_build_options(args: argparse.Namespace) -> None:
    if not os.path.isdir(args.data_dir):
        raise InvalidArgumentError(f'{args.data_dir} must be an existing directory')


def _normalize_build_write_data_options(args: argparse.Namespace) -> None:
    if args.command not in (
            'new-ansible', 'single', 'rebuild-single', 'multiple', 'changelog',
            'new-acd', 'build-single', 'build-multiple'):
        return

    if args.dest_data_dir is None:
        args.dest_data_dir = args.data_dir

    if not os.path.isdir(args.dest_data_dir):
        raise InvalidArgumentError(f'{args.dest_data_dir} must be an existing directory')


def _normalize_new_release_options(args: argparse.Namespace) -> None:
    flog = mlog.fields(func='_normalize_new_release_options')

    if args.command == 'new-acd':
        flog.warning('The new-acd command is deprecated.  Use `new-ansible` instead.')
        args.command = 'new-ansible'

    if args.command != 'new-ansible':
        return

    if args.pieces_file is None:
        args.pieces_file = DEFAULT_PIECES_FILE

    pieces_path = os.path.join(args.data_dir, args.pieces_file)
    if not os.path.isfile(pieces_path):
        raise InvalidArgumentError(f'The pieces file, {pieces_path}, must already'
                                   ' exist. It should contain one namespace.collection'
                                   ' per line')

    if args.build_file is None:
        basename = os.path.basename(os.path.splitext(args.pieces_file)[0])
        if args.ansible_version.major > 2:
            args.build_file = (
                f'{basename}-{args.ansible_version.major}.build'
            )
        else:
            args.build_file = (
                f'{basename}-{args.ansible_version.major}.{args.ansible_version.minor}.build'
            )


def _normalize_release_build_options(args: argparse.Namespace) -> None:
    flog = mlog.fields(func='_normalize_release_build_options')

    if args.command == 'build-single':
        flog.warning('The build-single command is deprecated.  Use `single` instead.')
        args.command = 'single'

    if args.command == 'build-multiple':
        flog.warning('The build-multiple command is deprecated.  Use `multiple` instead.')
        args.command = 'multiple'

    if args.command not in ('single', 'multiple', 'rebuild-single'):
        return

    compat_version_part = (
        f'{args.ansible_version.major}' if args.ansible_version.major > 2
        else f'{args.ansible_version.major}.{args.ansible_version.minor}'
    )

    if args.build_file is None:
        args.build_file = DEFAULT_FILE_BASE + f'-{compat_version_part}.build'

    build_filename = os.path.join(args.data_dir, args.build_file)
    if not os.path.isfile(build_filename):
        raise InvalidArgumentError(f'The build file, {build_filename} must already exist.'
                                   ' It should contains one namespace.collection and range'
                                   ' of versions per line')

    if args.deps_file is None:
        version_suffix = f'-{compat_version_part}'
        basename = os.path.basename(os.path.splitext(args.build_file)[0])
        if basename.endswith(version_suffix):
            basename = basename[:-len(version_suffix)]

        args.deps_file = f'{basename}-{args.ansible_version}.deps'

    if args.command in ('single', 'multiple'):
        if not os.path.isdir(args.sdist_dir):
            raise InvalidArgumentError(f'{args.sdist_dir} must be an existing directory')


def _normalize_release_rebuild_options(args: argparse.Namespace) -> None:
    if args.command not in ('rebuild-single', ):
        return

    deps_filename = os.path.join(args.data_dir, args.deps_file)
    if not os.path.isfile(deps_filename):
        raise InvalidArgumentError(f'The dependency file, {deps_filename} must already exist.')


def _normalize_collection_build_options(args: argparse.Namespace) -> None:
    flog = mlog.fields(func='_normalize_collection_build_options')

    if args.command == 'build-collection':
        flog.warning('The build-collection command is deprecated.  Use `collection` instead.')
        args.command = 'collection'

    if args.command != 'collection':
        return

    if args.deps_file is None:
        args.deps_file = DEFAULT_FILE_BASE + f'{args.ansible_version}.deps'

    if not os.path.isdir(args.collection_dir):
        raise InvalidArgumentError(f'{args.collection_dir} must be an existing directory')


def parse_args(program_name: str, args: List[str]) -> argparse.Namespace:
    """
    Parse and coerce the command line arguments.

    :arg program_name: The name of the program
    :arg args: A list of the command line arguments
    :returns: A :python:`argparse.Namespace`
    :raises InvalidArgumentError: Whenever there's something wrong with the arguments.
    """
    build_parser = argparse.ArgumentParser(add_help=False)
    build_parser.add_argument('ansible_version', type=PypiVer,
                              help='The X.Y.Z version of Ansible that this will be for')
    build_parser.add_argument('--data-dir', default='.',
                              help='Directory to read .build and .deps files from')

    build_write_data_parser = argparse.ArgumentParser(add_help=False, parents=[build_parser])
    build_write_data_parser.add_argument('--dest-data-dir', default=None,
                                         help='Directory to write .build and .deps files to,'
                                         ' as well as changelog and porting guide if applicable.'
                                         '  Defaults to --data-dir')

    cache_parser = argparse.ArgumentParser(add_help=False)
    cache_parser.add_argument('--collection-cache', default=None,
                              help='Directory of cached collection tarballs.  Will be'
                              ' used if a collection tarball to be downloaded exists'
                              ' in here, and will be populated when downloading new'
                              ' tarballs.')

    build_step_parser = argparse.ArgumentParser(add_help=False)
    build_step_parser.add_argument('--build-file', default=None,
                                   help='File containing the list of collections with version'
                                   ' ranges.  This is considered to be relative to'
                                   ' --build-data-dir.  The default is'
                                   ' $DEFAULT_FILE_BASE-X.Y.build')
    build_step_parser.add_argument('--deps-file', default=None,
                                   help='File which will be written containing the list of'
                                   ' collections at versions which were included in this version'
                                   ' of Ansible.  This is considered to be relative to'
                                   ' --build-data-dir.  The default is'
                                   ' $BASENAME_OF_BUILD_FILE-X.Y.Z.deps')

    feature_freeze_parser = argparse.ArgumentParser(add_help=False)
    feature_freeze_parser.add_argument('--feature-frozen', action='store_true',
                                       help='If this is given, then do not allow collections whose'
                                       ' version implies there are new features.')

    parser = get_toplevel_parser(prog=program_name,
                                 description='Script to manage building Ansible')

    subparsers = parser.add_subparsers(title='Subcommands', dest='command',
                                       help='for help use antsibull-build SUBCOMMANDS -h')
    subparsers.required = True

    new_parser = subparsers.add_parser('new-ansible', parents=[build_write_data_parser],
                                       description='Generate a new build description from the'
                                       ' latest available versions of ansible-base and the'
                                       ' included collections')
    new_parser.add_argument('--pieces-file', default=None,
                            help='File containing a list of collections to include.  This is'
                            ' considered to be relative to --data-dir.  The default is'
                            f' {DEFAULT_PIECES_FILE}')
    new_parser.add_argument('--build-file', default=None,
                            help='File which will be written which contains the list'
                            ' of collections with version ranges.  This is considered to be'
                            ' relative to --dest-data-dir.  The default is'
                            ' $BASENAME_OF_PIECES_FILE-X.Y.build')

    build_single_parser = subparsers.add_parser('single',
                                                parents=[build_write_data_parser, cache_parser,
                                                         build_step_parser, feature_freeze_parser],
                                                description='Build a single-file Ansible')
    build_single_parser.add_argument('--sdist-dir', default='.',
                                     help='Directory to write the generated sdist tarball to')
    build_single_parser.add_argument('--debian', action='store_true',
                                     help='Include Debian/Ubuntu packaging files in'
                                     ' the resulting output directory')

    rebuild_single_parser = subparsers.add_parser('rebuild-single',
                                                  parents=[build_write_data_parser, cache_parser,
                                                           build_step_parser],
                                                  description='Rebuild a single-file Ansible from'
                                                              ' a dependency file')
    rebuild_single_parser.add_argument('--sdist-dir', default='.',
                                       help='Directory to write the generated sdist tarball to')
    rebuild_single_parser.add_argument('--debian', action='store_true',
                                       help='Include Debian/Ubuntu packaging files in'
                                       ' the resulting output directory')

    build_multiple_parser = subparsers.add_parser('multiple',
                                                  parents=[build_write_data_parser, cache_parser,
                                                           build_step_parser,
                                                           feature_freeze_parser],
                                                  description='Build a multi-file Ansible')
    build_multiple_parser.add_argument('--sdist-dir', default='.',
                                       help='Directory to write the generated sdist tarballs to')

    collection_parser = subparsers.add_parser('collection',
                                              parents=[build_parser],
                                              description='Build a collection which will'
                                              ' install Ansible')
    collection_parser.add_argument('--deps-file', default=None,
                                   help='File which contains the list of collections and'
                                   ' versions which were included in this version of Ansible.'
                                   '  This is considered to be relative to --data-dir.'
                                   f'  The default is {DEFAULT_FILE_BASE}-X.Y.Z.deps')
    build_parser.add_argument('--collection-dir', default='.',
                              help='Directory to write collection to')

    subparsers.add_parser('changelog',
                          parents=[build_write_data_parser, cache_parser],
                          description='Build the Ansible changelog')

    # Backwards compat
    subparsers.add_parser('new-acd', add_help=False, parents=[new_parser])
    subparsers.add_parser('build-single', add_help=False, parents=[build_single_parser])
    subparsers.add_parser('build-multiple', add_help=False, parents=[build_multiple_parser])
    subparsers.add_parser('build-collection', add_help=False, parents=[collection_parser])

    args: argparse.Namespace = parser.parse_args(args)

    # Validation and coercion
    normalize_toplevel_options(args)
    _normalize_build_options(args)
    _normalize_build_write_data_options(args)
    _normalize_new_release_options(args)
    _normalize_release_build_options(args)
    _normalize_release_rebuild_options(args)
    _normalize_collection_build_options(args)

    return args


def run(args: List[str]) -> int:
    """
    Run the program.

    :arg args: A list of command line arguments.  Typically :python:`sys.argv`.
    :returns: A program return code.  0 for success, integers for any errors.  These are documented
        in :func:`main`.
    """
    flog = mlog.fields(func='run')
    flog.fields(raw_args=args).info('Enter')

    program_name = os.path.basename(args[0])
    try:
        args: argparse.Namespace = parse_args(program_name, args[1:])
    except InvalidArgumentError as e:
        print(e)
        return 2

    cfg = load_config(args.config_file)
    flog.fields(config=cfg).info('Config loaded')

    context_data = app_context.create_contexts(args=args, cfg=cfg)
    with app_context.app_and_lib_context(context_data) as (app_ctx, dummy_):
        twiggy.dict_config(app_ctx.logging_cfg.dict())
        flog.debug('Set logging config')

        return ARGS_MAP[args.command]()


def main() -> int:
    """
    Entrypoint called from the script.

    console_scripts call functions which take no parameters.  However, it's hard to test a function
    which takes no parameters so this function lightly wraps :func:`run`, which actually does the
    heavy lifting.

    :returns: A program return code.

    Return codes:
        :0: Success
        :1: Unhandled error.  See the Traceback for more information.
        :2: There was a problem with the command line arguments
        :3: version in an input file does not match with the version specified on the command line
        :4: Needs to be run on a newer version of Python
    """
    return run(sys.argv)
